import os
import pickle
import stat
import sys
import warnings
from pathlib import Path

import pytest
from fsspec import __version__ as fsspec_version
from fsspec import filesystem
from fsspec import get_filesystem_class
from packaging.version import Version
from pathlib_abc import PathParser
from pathlib_abc import vfspath

from upath import UnsupportedOperation
from upath import UPath
from upath._protocol import get_upath_protocol
from upath._stat import UPathStatResult
from upath.tests.utils import posixify
from upath.types import StatResultType


class JoinablePathTests:
    """Tests for JoinablePath interface.

    These tests verify pure path operations that don't require filesystem access:
    - Path parsing and components (parts, parents, name, stem, suffix, etc.)
    - Path manipulation (with_name, with_suffix, with_stem, joinpath, etc.)
    - Path comparison and hashing
    - Serialization (pickling)
    - URI handling
    """

    path: UPath

    def test_is_correct_class(self):
        raise NotImplementedError("must override")

    def test_parser(self):
        parser = self.path.parser
        assert isinstance(parser, PathParser)
        assert isinstance(parser.sep, str)
        assert parser.altsep is None or isinstance(parser.altsep, str)
        assert callable(parser.split)
        assert callable(parser.splitext)
        assert callable(parser.normcase)

    def test_with_segments(self):
        p = self.path.with_segments(self.path.__vfspath__(), "folder", "file.txt")
        assert p.parts[-2:] == ("folder", "file.txt")
        assert type(p) is type(self.path)

    def test___vfspath__(self):
        assert hasattr(self.path, "__vfspath__")
        assert callable(self.path.__vfspath__)
        str_path = vfspath(self.path)
        assert isinstance(str_path, str)

    def test_anchor(self):
        anchor = self.path.anchor
        assert isinstance(anchor, str)
        assert anchor == self.path.drive + self.path.root

    def test_is_absolute(self):
        assert self.path.is_absolute() is True

    def test_parents(self):
        p = self.path.joinpath("folder1", "file1.txt")
        assert p.parents[0] == p.parent
        assert p.parents[1] == p.parent.parent
        assert p.parents[0].name == "folder1"
        assert p.parents[1].name == self.path.name

    def test_with_name(self):
        path = self.path / "file.txt"
        path = path.with_name("file.zip")
        assert path.name == "file.zip"

    def test_with_suffix(self):
        path = self.path / "file.txt"
        path = path.with_suffix(".zip")
        assert path.suffix == ".zip"

    def test_suffix(self):
        path = self.path / "no_suffix"
        assert path.suffix == ""
        path = self.path / "file.txt"
        assert path.suffix == ".txt"
        path = self.path / "archive.tar.gz"
        assert path.suffix == ".gz"

    def test_suffixes(self):
        path = self.path / "no_suffix"
        assert path.suffixes == []
        path = self.path / "file.txt"
        assert path.suffixes == [".txt"]
        path = self.path / "archive.tar.gz"
        assert path.suffixes == [".tar", ".gz"]

    def test_with_stem(self):
        if sys.version_info < (3, 9):
            pytest.skip("with_stem only available on py3.9+")
        path = self.path / "file.txt"
        path = path.with_stem("document")
        assert path.stem == "document"

    def test_repr_after_with_name(self):
        p = self.path.joinpath("file.txt").with_name("file.zip")
        assert "file.zip" in repr(p)

    def test_repr_after_with_suffix(self):
        p = self.path.joinpath("file.txt").with_suffix(".zip")
        assert "file.zip" in repr(p)

    def test_child_path(self):
        path_str = self.path.__vfspath__()
        path_a = UPath(
            path_str, "folder", protocol=self.path.protocol, **self.path.storage_options
        )
        path_b = self.path / "folder"

        assert str(path_a) == str(path_b)
        assert path_a.root == path_b.root
        assert path_a.drive == path_b.drive

    def test_copy_path(self):
        path = self.path
        copy_path = UPath(path)

        assert type(path) is type(copy_path)
        assert str(path) == str(copy_path)
        assert path.drive == copy_path.drive
        assert path.root == copy_path.root
        assert path.parts == copy_path.parts
        assert path.storage_options == copy_path.storage_options

    def test_pickling(self):
        path = self.path
        pickled_path = pickle.dumps(path)
        recovered_path = pickle.loads(pickled_path)

        assert type(path) is type(recovered_path)
        assert str(path) == str(recovered_path)
        assert path.storage_options == recovered_path.storage_options

    def test_pickling_child_path(self):
        path = self.path / "subfolder" / "subsubfolder"
        pickled_path = pickle.dumps(path)
        recovered_path = pickle.loads(pickled_path)

        assert type(path) is type(recovered_path)
        assert str(path) == str(recovered_path)
        assert path.drive == recovered_path.drive
        assert path.root == recovered_path.root
        assert path.parts == recovered_path.parts
        assert path.storage_options == recovered_path.storage_options

    def test_as_uri(self):
        # test that we can reconstruct the path from the uri
        p0 = self.path
        uri = p0.as_uri()
        p1 = UPath(uri, **p0.storage_options)
        assert p0 == p1

    def test_protocol(self):
        protocol = self.path.protocol
        fs_cls = get_filesystem_class(protocol)
        protocols = [p] if isinstance((p := fs_cls.protocol), str) else p
        print(protocol, protocols)
        assert protocol in protocols

    def test_hashable(self):
        assert hash(self.path)

    def test_storage_options_dont_affect_hash(self):
        cls = type(self.path)
        p0 = cls(str(self.path), test_extra=1, **self.path.storage_options)
        p1 = cls(str(self.path), test_extra=2, **self.path.storage_options)
        assert hash(p0) == hash(p1)

    def test_eq(self):
        cls = type(self.path)
        p0 = cls(str(self.path), test_extra=1, **self.path.storage_options)
        p1 = cls(str(self.path), test_extra=1, **self.path.storage_options)
        p2 = cls(str(self.path), test_extra=2, **self.path.storage_options)
        assert p0 == p1
        assert p0 != p2
        assert p1 != p2

    def test_relative_to(self):
        base = self.path
        child = self.path / "folder1" / "file1.txt"
        relative = child.relative_to(base)
        assert str(relative) == "folder1/file1.txt"

    def test_is_relative_to(self):
        base = self.path
        child = self.path / "folder1" / "file1.txt"
        other = UPath("/some/other/path")

        assert child.is_relative_to(base) is True
        assert base.is_relative_to(child) is False
        assert child.is_relative_to(other) is False

    def test_full_match(self):
        p = self.path / "folder" / "file.txt"
        assert p.full_match("**/*") is True
        assert p.full_match("**/*.txt") is True
        assert p.full_match("*.doesnotexist") is False

    def test_trailing_slash_joinpath_is_identical(self):
        # setup
        cls = type(self.path)
        protocol = self.path.protocol
        path = self.path.path
        sopts = self.path.storage_options
        if not path:
            path = "something"
            path_with_slash = "something/"
        elif path.endswith("/"):
            path_with_slash = path
            path = path.removeprefix("/")
        else:
            path_with_slash = path + "/"
        key = "key/"

        # test
        a = cls(path_with_slash + key, protocol=protocol, **sopts)
        b = cls(path_with_slash, key, protocol=protocol, **sopts)
        c = cls(path_with_slash, protocol=protocol, **sopts).joinpath(key)
        d = cls(path_with_slash, protocol=protocol, **sopts) / key
        assert a.path == b.path == c.path == d.path

    def test_trailing_slash_is_stripped(self):
        has_meaningful_trailing_slash = getattr(
            self.path.parser, "has_meaningful_trailing_slash", False
        )
        if has_meaningful_trailing_slash:
            assert not self.path.joinpath("key").path.endswith("/")
            assert self.path.joinpath("key/").path.endswith("/")
        else:
            assert not self.path.joinpath("key").path.endswith("/")
            assert not self.path.joinpath("key/").path.endswith("/")

    def test_parents_are_absolute(self):
        # this is a cross implementation compatible way to ensure that
        # the path representing the root is absolute
        is_absolute = [p.is_absolute() for p in self.path.parents]
        assert all(is_absolute)

    def test_parents_end_at_anchor(self):
        p = self.path.joinpath("folder1", "file1.txt")
        assert p.parents[-1].path == posixify(p.anchor)

    def test_anchor_is_its_own_parent(self):
        p = self.path.joinpath("folder1", "file1.txt")
        p0 = p.parents[-1]
        assert p0.path == posixify(p.anchor)
        assert p0.parent.path == posixify(p.anchor)

    def test_private_url_attr_in_sync(self):
        p = self.path
        p1 = self.path.joinpath("c")
        p2 = self.path / "c"
        assert p1._url == p2._url
        assert p1._url != p._url
        assert p1.protocol == p2.protocol


class ReadablePathTests:
    """Tests for ReadablePath interface.

    These tests verify operations that read from the filesystem:
    - File/directory existence and type checks (exists, is_dir, is_file, etc.)
    - File metadata (stat, info)
    - Reading file contents (read_bytes, read_text, open for reading)
    - Directory listing (iterdir, glob, rglob)
    - Copy operations (read source)
    """

    path: UPath

    @pytest.fixture(autouse=True)
    def path_file(self, path):
        self.path_file = self.path.joinpath("file1.txt")

    def test_storage_options_match_fsspec(self):
        storage_options = self.path.storage_options
        assert storage_options == self.path.fs.storage_options

    def test_stat(self):
        stat_ = self.path.stat()

        # for debugging os.stat_result compatibility
        attrs = {attr for attr in dir(stat_) if attr.startswith("st_")}
        print(attrs)

        assert isinstance(stat_, StatResultType)
        assert len(tuple(stat_)) == os.stat_result.n_sequence_fields

        with warnings.catch_warnings():
            warnings.simplefilter("error")

            for idx in range(os.stat_result.n_sequence_fields):
                assert isinstance(stat_[idx], int)
            for attr in UPathStatResult._fields + UPathStatResult._fields_extra:
                assert hasattr(stat_, attr)

    def test_stat_dir_st_mode(self):
        base = self.path.stat()  # base folder
        assert stat.S_ISDIR(base.st_mode)

    def test_stat_file_st_mode(self):
        file1 = self.path_file.stat()
        assert stat.S_ISREG(file1.st_mode)

    def test_stat_st_size(self):
        file1 = self.path_file.stat()
        assert file1.st_size == 11

    @pytest.mark.parametrize(
        "url, expected", [("file1.txt", True), ("fakefile.txt", False)]
    )
    def test_exists(self, url, expected):
        path = self.path.joinpath(url)
        assert path.exists() == expected

    def test_expanduser(self):
        assert self.path.expanduser() == self.path

    @pytest.mark.parametrize(
        "pattern",
        (
            "*.txt",
            "*",
            pytest.param(
                "**/*.txt",
                marks=(
                    pytest.mark.xfail(reason="requires fsspec>=2023.9.0")
                    if Version(fsspec_version) < Version("2023.9.0")
                    else ()
                ),
            ),
        ),
    )
    def test_glob(self, pathlib_base, pattern):
        mock_glob = list(self.path.glob(pattern))
        path_glob = list(pathlib_base.glob(pattern))

        _mock_start = len(self.path.parts)
        mock_glob_normalized = sorted(
            [tuple(filter(None, a.parts[_mock_start:])) for a in mock_glob]
        )
        _path_start = len(pathlib_base.parts)
        path_glob_normalized = sorted([a.parts[_path_start:] for a in path_glob])

        print(mock_glob_normalized, path_glob_normalized)
        assert mock_glob_normalized == path_glob_normalized

    def test_is_dir(self):
        assert self.path.is_dir()

        path = self.path / "file1.txt"
        assert not path.is_dir()
        assert not (self.path / "not-existing-dir").is_dir()

    def test_is_file(self):
        path_exists = self.path / "file1.txt"
        assert path_exists.is_file()
        assert not (self.path / "not-existing-file.txt").is_file()

    def test_is_mount(self):
        try:
            self.path.is_mount()
        except UnsupportedOperation:
            pytest.skip(f"is_mount() not supported for {type(self.path).__name__}")
        else:
            assert self.path.is_mount() is False

    def test_is_symlink(self):
        assert self.path.is_symlink() is False

    def test_is_socket(self):
        assert self.path.is_socket() is False

    def test_is_fifo(self):
        assert self.path.is_fifo() is False

    def test_is_block_device(self):
        assert self.path.is_block_device() is False

    def test_is_char_device(self):
        assert self.path.is_char_device() is False

    def test_iterdir(self, local_testdir):
        pl_path = Path(local_testdir)

        up_iter = list(self.path.iterdir())
        pl_iter = list(pl_path.iterdir())

        for x in up_iter:
            assert x.name != ""
            assert x.exists()

        assert len(up_iter) == len(pl_iter)
        assert {p.name for p in pl_iter} == {u.name for u in up_iter}

    def test_iterdir_parent_iteration(self):
        assert next(self.path.parent.iterdir()).exists()

    def test_iterdir2(self, local_testdir):
        pl_path = Path(local_testdir) / "folder1"

        up_iter = list((self.path / "folder1").iterdir())
        pl_iter = list(pl_path.iterdir())

        for x in up_iter:
            assert x.exists()

        assert len(up_iter) == len(pl_iter)
        assert {p.name for p in pl_iter} == {u.name for u in up_iter}

    def test_iterdir_trailing_slash(self):
        files_noslash = list(self.path.joinpath("folder1").iterdir())
        files_slash = list(self.path.joinpath("folder1/").iterdir())
        assert files_noslash == files_slash

    def test_lstat(self):
        with pytest.warns(UserWarning, match=r"[A-Za-z]+.stat"):
            st = self.path.lstat()
            assert st is not None

    def test_cwd(self):
        with pytest.raises(UnsupportedOperation):
            self.path.cwd()

    def test_home(self):
        with pytest.raises(UnsupportedOperation):
            self.path.home()

    def test_open(self):
        p = self.path_file
        with p.open(mode="r") as f:
            assert f.read() == "hello world"
        with p.open(mode="rb") as f:
            assert f.read() == b"hello world"

    def test_open_buffering(self):
        p = self.path_file
        p.open(buffering=-1)

    def test_open_block_size(self):
        p = self.path_file
        with p.open(mode="r", block_size=8192) as f:
            assert f.read() == "hello world"

    def test_open_errors(self):
        p = self.path_file
        with p.open(mode="r", encoding="ascii", errors="strict") as f:
            assert f.read() == "hello world"

    def test_read_bytes(self):
        mock = self.path.joinpath("file2.txt")
        assert mock.read_bytes() == b"hello world"

    def test_read_text(self):
        upath = self.path.joinpath("file1.txt")
        assert upath.read_text() == "hello world"

    def test_read_text_encoding(self):
        upath = self.path_file
        content = upath.read_text(encoding="utf-8")
        assert content == "hello world"

    def test_read_text_errors(self):
        upath = self.path_file
        content = upath.read_text(encoding="ascii", errors="strict")
        assert content == "hello world"

    def test_rglob(self, pathlib_base):
        pattern = "*.txt"
        result = [*self.path.rglob(pattern)]
        expected = [*pathlib_base.rglob(pattern)]
        assert len(result) == len(expected)

    def test_walk(self, local_testdir):
        def _raise(x):
            raise x

        # collect walk results from UPath
        upath_walk = []
        for dirpath, dirnames, filenames in self.path.walk(on_error=_raise):
            rel_dirpath = dirpath.relative_to(self.path)
            upath_walk.append((str(rel_dirpath), sorted(dirnames), sorted(filenames)))
        upath_walk.sort()

        # collect walk results using os.walk (compatible with Python 3.9+)
        os_walk = []
        for dirpath, dirnames, filenames in os.walk(local_testdir):
            rel_dirpath = os.path.relpath(dirpath, local_testdir)
            os_walk.append((rel_dirpath, sorted(dirnames), sorted(filenames)))
        os_walk.sort()

        assert upath_walk == os_walk

    def test_walk_top_down_false(self):
        def _raise(x):
            raise x

        # test walk with top_down=False returns directories after their contents
        paths_seen = []
        for dirpath, _, _ in self.path.walk(top_down=False, on_error=_raise):
            paths_seen.append(dirpath)

        # in bottom-up walk, parent directories should come after children
        for i, path in enumerate(paths_seen):
            for _, other in enumerate(paths_seen[i + 1 :], start=i + 1):
                # if path is a parent of other, path should come after other
                if other.is_relative_to(path) and other != path:
                    pytest.fail(f"In bottom-up walk, {path} should come after {other}")

    def test_samefile(self):
        f1 = self.path.joinpath("file1.txt")
        f2 = self.path.joinpath("file2.txt")

        assert f1.samefile(f2) is False
        assert f1.samefile(f2.path) is False
        assert f1.samefile(f1) is True
        assert f1.samefile(f1.path) is True

    def test_info(self):
        p0 = self.path.joinpath("file1.txt")
        p1 = self.path.joinpath("folder1")

        assert p0.info.exists() is True
        assert p0.info.is_file() is True
        assert p0.info.is_dir() is False
        assert p0.info.is_symlink() is False
        assert p1.info.exists() is True
        assert p1.info.is_file() is False
        assert p1.info.is_dir() is True
        assert p1.info.is_symlink() is False

    def test_copy_local(self, tmp_path: Path):
        target = UPath(tmp_path) / "target-file1.txt"

        source = self.path_file
        content = source.read_text()
        source.copy(target)
        assert target.exists()
        assert target.read_text() == content

    @pytest.mark.parametrize("target_type", [str, Path, UPath])
    def test_copy_into__file_to_str_tempdir(self, tmp_path: Path, target_type):
        tmp_path = tmp_path.joinpath("somewhere")
        tmp_path.mkdir()
        target_dir = target_type(tmp_path)
        assert isinstance(target_dir, target_type)

        source = self.path_file
        source.copy_into(target_dir)
        target = tmp_path.joinpath(source.name)

        assert target.exists()
        assert target.read_text() == source.read_text()

    @pytest.mark.parametrize("target_type", [str, Path, UPath])
    def test_copy_into__dir_to_str_tempdir(self, tmp_path: Path, target_type):
        tmp_path = tmp_path.joinpath("somewhere")
        tmp_path.mkdir()
        target_dir = target_type(tmp_path)
        assert isinstance(target_dir, target_type)

        source_dir = self.path.joinpath("folder1")
        assert source_dir.is_dir()
        source_dir.copy_into(target_dir)
        target = tmp_path.joinpath(source_dir.name)

        assert target.exists()
        assert target.is_dir()
        for item in source_dir.iterdir():
            target_item = target.joinpath(item.name)
            assert target_item.exists()
            if item.is_file():
                assert target_item.read_text() == item.read_text()

    def test_copy_into_local(self, tmp_path: Path):
        target_dir = UPath(tmp_path) / "target-dir"
        target_dir.mkdir()

        source = self.path_file
        content = source.read_text()
        source.copy_into(target_dir)
        target = target_dir / source.name
        assert target.exists()
        assert target.read_text() == content

    def test_copy_memory(self, clear_fsspec_memory_cache):
        target = UPath("memory:///target-file1.txt")
        source = self.path_file
        content = source.read_text()
        source.copy(target)
        assert target.exists()
        assert target.read_text() == content

    def test_copy_into_memory(self, clear_fsspec_memory_cache):
        target_dir = UPath("memory:///target-dir")
        target_dir.mkdir()

        source = self.path_file
        content = source.read_text()
        source.copy_into(target_dir)
        target = target_dir / source.name
        assert target.exists()
        assert target.read_text() == content

    def test_copy_exceptions(self, tmp_path: Path):
        source = self.path_file
        # target is a directory
        target = UPath(tmp_path) / "target-folder"
        target.mkdir()
        # FIXME: pytest.raises(IsADirectoryError) not working on Windows
        with pytest.raises(OSError):
            source.copy(target)
        # target parent does not exist
        target = UPath(tmp_path) / "nonexistent-dir" / "target-file1.txt"
        with pytest.raises(FileNotFoundError):
            source.copy(target)

    def test_copy_into_exceptions(self, tmp_path: Path):
        source = self.path_file
        # target is not a directory
        target_file = UPath(tmp_path) / "target-file.txt"
        target_file.write_text("content")
        # FIXME: pytest.raises(NotADirectoryError) not working on Windows
        with pytest.raises(OSError):
            source.copy_into(target_file)
        # target dir does not exist
        target_dir = UPath(tmp_path) / "nonexistent-dir"
        with pytest.raises(FileNotFoundError):
            source.copy_into(target_dir)

    def test_read_with_fsspec(self):
        p = self.path_file

        protocol = p.protocol
        storage_options = p.storage_options
        path = p.path

        fs = filesystem(protocol, **storage_options)
        with fs.open(path) as f:
            assert f.read() == b"hello world"

    def test_readlink(self):
        with pytest.raises(UnsupportedOperation):
            self.path.readlink()

    def test_group(self):
        with pytest.raises(UnsupportedOperation):
            self.path.group()

    def test_owner(self):
        with pytest.raises(UnsupportedOperation):
            self.path.owner()


# =============================================================================
# WritablePathTests: Tests for writable path operations
# =============================================================================


class _CommonWritablePathTests:

    SUPPORTS_EMPTY_DIRS = True

    path: UPath

    def test_chmod(self):
        with pytest.raises(NotImplementedError):
            self.path_file.chmod(777)

    def test_lchmod(self):
        with pytest.raises(UnsupportedOperation):
            self.path.lchmod(mode=0o777)

    def test_symlink_to(self):
        with pytest.raises(UnsupportedOperation):
            self.path_file.symlink_to("target")
        with pytest.raises(UnsupportedOperation):
            self.path.joinpath("link").symlink_to("target")

    def test_hardlink_to(self):
        with pytest.raises(UnsupportedOperation):
            self.path_file.symlink_to("target")
        with pytest.raises(UnsupportedOperation):
            self.path.joinpath("link").hardlink_to("target")


class NonWritablePathTests(_CommonWritablePathTests):

    def test_mkdir_raises(self):
        with pytest.raises(UnsupportedOperation):
            self.path.mkdir()

    def test_touch_raises(self):
        with pytest.raises(UnsupportedOperation):
            self.path.touch()

    def test_unlink(self):
        with pytest.raises(UnsupportedOperation):
            self.path.unlink()

    def test_write_bytes(self):
        with pytest.raises(UnsupportedOperation):
            self.path_file.write_bytes(b"abc")

    def test_write_text(self):
        with pytest.raises(UnsupportedOperation):
            self.path_file.write_text("abc")


class WritablePathTests(_CommonWritablePathTests):
    """Tests for WritablePath interface.

    These tests verify operations that write to the filesystem:
    - Creating directories (mkdir)
    - Creating files (touch)
    - Writing file contents (write_bytes, write_text)
    - Removing files/directories (unlink, rmdir)
    """

    def test_mkdir(self):
        new_dir = self.path.joinpath("new_dir")
        new_dir.mkdir()
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        assert new_dir.exists()

    def test_mkdir_exists_ok_true(self):
        new_dir = self.path.joinpath("new_dir_may_exists")
        new_dir.mkdir()
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        new_dir.mkdir(exist_ok=True)

    def test_mkdir_exists_ok_false(self):
        new_dir = self.path.joinpath("new_dir_may_not_exists")
        new_dir.mkdir()
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        with pytest.raises(FileExistsError):
            new_dir.mkdir(exist_ok=False)

    def test_mkdir_parents_true_exists_ok_true(self):
        new_dir = self.path.joinpath("parent", "new_dir_may_not_exist")
        new_dir.mkdir(parents=True)
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        new_dir.mkdir(parents=True, exist_ok=True)

    def test_mkdir_parents_true_exists_ok_false(self):
        new_dir = self.path.joinpath("parent", "new_dir_may_exist")
        new_dir.mkdir(parents=True)
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        with pytest.raises(FileExistsError):
            new_dir.mkdir(parents=True, exist_ok=False)

    def test_touch_exists_ok_false(self):
        f = self.path.joinpath("file1.txt")
        assert f.exists()
        with pytest.raises(FileExistsError):
            f.touch(exist_ok=False)

    def test_touch_exists_ok_true(self):
        f = self.path.joinpath("file1.txt")
        assert f.exists()
        data = f.read_text()
        f.touch(exist_ok=True)
        assert f.read_text() == data

    def test_touch(self):
        path = self.path.joinpath("test_touch.txt")
        assert not path.exists()
        path.touch()
        assert path.exists()

    def test_touch_unlink(self):
        path = self.path.joinpath("test_touch.txt")
        path.touch()
        assert path.exists()
        path.unlink()
        assert not path.exists()

        # should raise FileNotFoundError since file is missing
        with pytest.raises(FileNotFoundError):
            path.unlink()

        # file doesn't exists, but missing_ok is True
        path.unlink(missing_ok=True)

    def test_write_bytes(self, pathlib_base):
        fn = "test_write_bytes.txt"
        s = b"hello_world"
        path = self.path.joinpath(fn)
        path.write_bytes(s)
        assert path.read_bytes() == s

    def test_write_text(self, pathlib_base):
        fn = "test_write_text.txt"
        s = "hello_world"
        path = self.path.joinpath(fn)
        path.write_text(s)
        assert path.read_text() == s

    def test_write_text_encoding(self):
        fn = "test_write_text_enc.txt"
        s = "hello_world"
        path = self.path.joinpath(fn)
        path.write_text(s, encoding="utf-8")
        assert path.read_text(encoding="utf-8") == s

    def test_write_text_errors(self):
        fn = "test_write_text_errors.txt"
        s = "hello_world"
        path = self.path.joinpath(fn)
        path.write_text(s, encoding="ascii", errors="strict")
        assert path.read_text(encoding="ascii") == s


class ReadWritePathTests:
    """Tests requiring both ReadablePath and WritablePath interfaces.

    These tests verify operations that need both read and write access:
    - Rename/move operations
    - File system setup/teardown
    - Operations that verify write results by reading
    - rmdir operations
    """

    SUPPORTS_EMPTY_DIRS = True

    path: UPath

    def test_rename(self):
        p_source = self.path.joinpath("file1.txt")
        p_target = self.path.joinpath("file1_renamed.txt")

        p_moved = p_source.rename(p_target)
        assert p_target == p_moved
        assert not p_source.exists()
        assert p_moved.exists()

        p_revert = p_moved.rename(p_source)
        assert p_revert == p_source
        assert not p_moved.exists()
        assert p_revert.exists()

    @pytest.fixture
    def supports_cwd(self):
        # intentionally called on the instance to support ProxyUPath().cwd()
        try:
            self.path.cwd()
        except UnsupportedOperation:
            return False
        else:
            return True

    @pytest.mark.parametrize(
        "target_factory",
        [
            lambda obj, name: name,
            lambda obj, name: UPath(name),
            lambda obj, name: Path(name),
            lambda obj, name: obj.joinpath(name).relative_to(obj),
        ],
        ids=[
            "str_relative",
            "plain_upath_relative",
            "plain_path_relative",
            "self_upath_relative",
        ],
    )
    def test_rename_with_target_relative(
        self, request, monkeypatch, supports_cwd, target_factory, tmp_path
    ):
        source = self.path.joinpath("folder1/file2.txt")
        target = target_factory(self.path, "file2_renamed.txt")

        source_text = source.read_text()
        if supports_cwd:
            cid = request.node.callspec.id
            cwd = tmp_path.joinpath(cid)
            cwd.mkdir(parents=True, exist_ok=True)
            monkeypatch.chdir(cwd)

            t = source.rename(target)
            assert (t.protocol == UPath(target).protocol) or UPath(
                target
            ).protocol == ""
            assert (t.path == UPath(target).path) or (
                t.path == UPath(target).absolute().path
            )
            assert t.exists()
            assert t.read_text() == source_text

        else:
            with pytest.raises(UnsupportedOperation):
                source.rename(target)

    @pytest.mark.parametrize(
        "target_factory",
        [
            lambda obj, name: obj.joinpath(name).absolute().as_posix(),
            lambda obj, name: UPath(obj.absolute().joinpath(name).path),
            lambda obj, name: Path(obj.absolute().joinpath(name).path),
            lambda obj, name: obj.absolute().joinpath(name),
        ],
        ids=[
            "str_absolute",
            "plain_upath_absolute",
            "plain_path_absolute",
            "self_upath_absolute",
        ],
    )
    def test_rename_with_target_absolute(self, target_factory):
        from upath._chain import Chain
        from upath._chain import FSSpecChainParser

        source = self.path.joinpath("folder1/file2.txt")
        target = target_factory(self.path, "file2_renamed.txt")

        source_text = source.read_text()
        t = source.rename(target)
        assert get_upath_protocol(target) in {t.protocol, ""}
        assert t.path == Chain.from_list(
            FSSpecChainParser().unchain(str(target))
        ).active_path.replace("\\", "/")
        assert t.exists()
        assert t.read_text() == source_text

    def test_replace(self):
        pass

    def test_resolve(self):
        pass

    def test_rmdir_no_dir(self):
        p = self.path.joinpath("file1.txt")
        with pytest.raises(NotADirectoryError):
            p.rmdir()

    def test_iterdir_no_dir(self):
        p = self.path.joinpath("file1.txt")
        assert p.is_file()
        with pytest.raises(NotADirectoryError):
            _ = list(p.iterdir())

    def test_rmdir_not_empty(self):
        p = self.path.joinpath("folder1")
        with pytest.raises(OSError, match="not empty"):
            p.rmdir(recursive=False)

    def test_fsspec_compat(self):
        fs = self.path.fs
        content = b"a,b,c\n1,2,3\n4,5,6"

        upath1 = self.path / "output1.csv"
        p1 = upath1.path
        upath1.write_bytes(content)
        assert fs._fs_token == upath1.fs._fs_token
        if fs.cachable:  # codespell:ignore cachable
            assert fs is upath1.fs
        with fs.open(p1) as f:
            assert f.read() == content
        upath1.unlink()

        # write with fsspec, read with upath
        upath2 = self.path / "output2.csv"
        p2 = upath2.path
        if fs.cachable:  # codespell:ignore cachable
            assert fs is upath2.fs
        with fs.open(p2, "wb") as f:
            f.write(content)
        assert upath2.read_bytes() == content
        upath2.unlink()

    def test_move_local(self, tmp_path: Path):
        target = UPath(tmp_path) / "target-file1.txt"

        source = self.path / "file1.txt"
        content = source.read_text()
        source.move(target)
        assert target.exists()
        assert target.read_text() == content
        assert not source.exists()

    def test_move_into_local(self, tmp_path: Path):
        target_dir = UPath(tmp_path) / "target-dir"
        target_dir.mkdir()

        source = self.path / "file1.txt"
        content = source.read_text()
        source.move_into(target_dir)
        target = target_dir / "file1.txt"
        assert target.exists()
        assert target.read_text() == content
        assert not source.exists()

    def test_move_memory(self, clear_fsspec_memory_cache):
        target = UPath("memory:///target-file1.txt")
        source = self.path / "file1.txt"
        content = source.read_text()
        source.move(target)
        assert target.exists()
        assert target.read_text() == content
        assert not source.exists()

    def test_move_into_memory(self, clear_fsspec_memory_cache):
        target_dir = UPath("memory:///target-dir")
        target_dir.mkdir()

        source = self.path / "file1.txt"
        content = source.read_text()
        source.move_into(target_dir)
        target = target_dir / "file1.txt"
        assert target.exists()
        assert target.read_text() == content
        assert not source.exists()

    def prepare_file_system(self):
        self.make_top_folder()
        self.make_test_files()

    def make_top_folder(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def make_test_files(self):
        folder1 = self.path.joinpath("folder1")
        folder1.mkdir(exist_ok=True)
        folder1_files = ["file1.txt", "file2.txt"]
        for f in folder1_files:
            p = folder1.joinpath(f)
            p.touch()
            p.write_text(f)

        file1 = self.path.joinpath("file1.txt")
        file1.touch()
        file1.write_text("hello world")
        file2 = self.path.joinpath("file2.txt")
        file2.touch()
        file2.write_bytes(b"hello world")


class BaseTests(
    JoinablePathTests,
    ReadablePathTests,
    WritablePathTests,
    ReadWritePathTests,
):
    """Comprehensive test suite combining all path operation tests.

    This class composes all the individual test suites for testing UPath
    implementations that support full read/write functionality. For UPath
    subclasses with limited functionality (e.g., read-only), use the
    appropriate subset of test classes:

    - JoinablePathTests: Pure path operations (no I/O)
    - ReadablePathTests: Read-only operations
    - WritablePathTests: Write-only operations
    - ReadWritePathTests: Operations requiring both read and write

    Example usage for a read-only UPath:

        class TestMyReadOnlyPath(JoinablePathTests, ReadablePathTests):
            @pytest.fixture(autouse=True)
            def setup(self, ...):
                self.path = MyReadOnlyUPath(...)
    """

    SUPPORTS_EMPTY_DIRS = True

    path: UPath



"""vendoring everything needed to run test_pathlib.py

https://github.com/python/cpython/tree/10bf2cd404320252ef162d5699cb7ce52a970d44

from test.support import import_helper
from test.support import set_recursion_limit
from test.support import is_emscripten, is_wasi
from test.support import os_helper
from test.support.os_helper import TESTFN, FakePath
"""
import os
import stat
import string
import sys
import collections.abc
import contextlib
import time
import warnings
from types import SimpleNamespace

import pytest
from pytest import importorskip

import_helper = SimpleNamespace(import_module=importorskip)
import_module = importorskip


@contextlib.contextmanager
def set_recursion_limit(limit):
    """Temporarily change the recursion limit."""
    original_limit = sys.getrecursionlimit()
    try:
        sys.setrecursionlimit(limit)
        yield
    finally:
        sys.setrecursionlimit(original_limit)


is_emscripten = sys.platform == "emscripten"
is_wasi = sys.platform == "wasi"


class FakePath:
    """Simple implementation of the path protocol."""

    def __init__(self, path):
        self.path = path

    def __repr__(self):
        return f"<FakePath {self.path!r}>"

    def __fspath__(self):
        if (
            isinstance(self.path, BaseException)
            or isinstance(self.path, type)
            and issubclass(self.path, BaseException)
        ):
            raise self.path
        else:
            return self.path


class EnvironmentVarGuard(collections.abc.MutableMapping):

    """Class to help protect the environment variable properly.  Can be used as
    a context manager."""

    def __init__(self):
        self._environ = os.environ
        self._changed = {}

    def __getitem__(self, envvar):
        return self._environ[envvar]

    def __setitem__(self, envvar, value):
        # Remember the initial value on the first access
        if envvar not in self._changed:
            self._changed[envvar] = self._environ.get(envvar)
        self._environ[envvar] = value

    def __delitem__(self, envvar):
        # Remember the initial value on the first access
        if envvar not in self._changed:
            self._changed[envvar] = self._environ.get(envvar)
        if envvar in self._environ:
            del self._environ[envvar]

    def keys(self):
        return self._environ.keys()

    def __iter__(self):
        return iter(self._environ)

    def __len__(self):
        return len(self._environ)

    def set(self, envvar, value):
        self[envvar] = value

    def unset(self, envvar):
        del self[envvar]

    def copy(self):
        # We do what os.environ.copy() does.
        return dict(self)

    def __enter__(self):
        return self

    def __exit__(self, *ignore_exc):
        for k, v in self._changed.items():
            if v is None:
                if k in self._environ:
                    del self._environ[k]
            else:
                self._environ[k] = v
        os.environ = self._environ


# Filename used for testing
TESTFN_ASCII = "@test"

# Disambiguate TESTFN for parallel testing, while letting it remain a valid
# module name.
TESTFN_ASCII = "{}_{}_tmp".format(TESTFN_ASCII, os.getpid())


# FS_NONASCII: non-ASCII character encodable by os.fsencode(),
# or an empty string if there is no such character.
FS_NONASCII = ""
for character in (
    # First try printable and common characters to have a readable filename.
    # For each character, the encoding list are just example of encodings able
    # to encode the character (the list is not exhaustive).
    # U+00E6 (Latin Small Letter Ae): cp1252, iso-8859-1
    "\u00E6",
    # U+0130 (Latin Capital Letter I With Dot Above): cp1254, iso8859_3
    "\u0130",
    # U+0141 (Latin Capital Letter L With Stroke): cp1250, cp1257
    "\u0141",
    # U+03C6 (Greek Small Letter Phi): cp1253
    "\u03C6",
    # U+041A (Cyrillic Capital Letter Ka): cp1251
    "\u041A",
    # U+05D0 (Hebrew Letter Alef): Encodable to cp424
    "\u05D0",
    # U+060C (Arabic Comma): cp864, cp1006, iso8859_6, mac_arabic
    "\u060C",
    # U+062A (Arabic Letter Teh): cp720
    "\u062A",
    # U+0E01 (Thai Character Ko Kai): cp874
    "\u0E01",
    # Then try more "special" characters. "special" because they may be
    # interpreted or displayed differently depending on the exact locale
    # encoding and the font.
    # U+00A0 (No-Break Space)
    "\u00A0",
    # U+20AC (Euro Sign)
    "\u20AC",
):
    try:
        # If Python is set up to use the legacy 'mbcs' in Windows,
        # 'replace' error mode is used, and encode() returns b'?'
        # for characters missing in the ANSI codepage
        if os.fsdecode(os.fsencode(character)) != character:
            raise UnicodeError
    except UnicodeError:
        pass
    else:
        FS_NONASCII = character
        break


# TESTFN_UNDECODABLE is a filename (bytes type) that should *not* be able to be
# decoded from the filesystem encoding (in strict mode). It can be None if we
# cannot generate such filename (ex: the latin1 encoding can decode any byte
# sequence). On UNIX, TESTFN_UNDECODABLE can be decoded by os.fsdecode() thanks
# to the surrogateescape error handler (PEP 383), but not from the filesystem
# encoding in strict mode.
TESTFN_UNDECODABLE = None
for name in (
    # b'\xff' is not decodable by os.fsdecode() with code page 932. Windows
    # accepts it to create a file or a directory, or don't accept to enter to
    # such directory (when the bytes name is used). So test b'\xe7' first:
    # it is not decodable from cp932.
    b"\xe7w\xf0",
    # undecodable from ASCII, UTF-8
    b"\xff",
    # undecodable from iso8859-3, iso8859-6, iso8859-7, cp424, iso8859-8, cp856
    # and cp857
    b"\xae\xd5"
    # undecodable from UTF-8 (UNIX and Mac OS X)
    b"\xed\xb2\x80",
    b"\xed\xb4\x80",
    # undecodable from shift_jis, cp869, cp874, cp932, cp1250, cp1251, cp1252,
    # cp1253, cp1254, cp1255, cp1257, cp1258
    b"\x81\x98",
):
    try:
        name.decode(sys.getfilesystemencoding())
    except UnicodeDecodeError:
        try:
            name.decode(
                sys.getfilesystemencoding(), sys.getfilesystemencodeerrors()
            )
        except UnicodeDecodeError:
            continue
        TESTFN_UNDECODABLE = os.fsencode(TESTFN_ASCII) + name
        break

if FS_NONASCII:
    TESTFN_NONASCII = TESTFN_ASCII + FS_NONASCII
else:
    TESTFN_NONASCII = None
TESTFN = TESTFN_NONASCII or TESTFN_ASCII


_can_symlink = None


def can_symlink():
    global _can_symlink
    if _can_symlink is not None:
        return _can_symlink
    # WASI / wasmtime prevents symlinks with absolute paths, see man
    # openat2(2) RESOLVE_BENEATH. Almost all symlink tests use absolute
    # paths. Skip symlink tests on WASI for now.
    src = os.path.abspath(TESTFN)
    symlink_path = src + "can_symlink"
    try:
        os.symlink(src, symlink_path)
        can = True
    except (OSError, NotImplementedError, AttributeError):
        can = False
    else:
        os.remove(symlink_path)
    _can_symlink = can
    return can


def skip_unless_symlink(test):
    """Skip decorator for tests that require functional symlink"""
    ok = can_symlink()
    msg = "Requires functional symlink implementation"
    return test if ok else pytest.mark.skip(msg)(test)


_can_chmod = None


def can_chmod():
    global _can_chmod
    if _can_chmod is not None:
        return _can_chmod
    if not hasattr(os, "chown"):
        _can_chmod = False
        return _can_chmod
    try:
        with open(TESTFN, "wb") as _:
            try:
                os.chmod(TESTFN, 0o777)
                mode1 = os.stat(TESTFN).st_mode
                os.chmod(TESTFN, 0o666)
                mode2 = os.stat(TESTFN).st_mode
            except OSError:
                can = False
            else:
                can = stat.S_IMODE(mode1) != stat.S_IMODE(mode2)
    finally:
        unlink(TESTFN)
    _can_chmod = can
    return can


def skip_unless_working_chmod(test):
    """Skip tests that require working os.chmod()

    WASI SDK 15.0 cannot change file mode bits.
    """
    ok = can_chmod()
    msg = "requires working os.chmod()"
    return test if ok else pytest.mark.skip(msg)(test)


def unlink(filename):
    try:
        _unlink(filename)
    except (FileNotFoundError, NotADirectoryError):
        pass


def _force_run(path, func, *args):
    try:
        return func(*args)
    except FileNotFoundError:
        # chmod() won't fix a missing file.
        raise
    except OSError:
        os.chmod(path, stat.S_IRWXU)
        return func(*args)


if sys.platform.startswith("win"):
    def _waitfor(func, pathname, waitall=False):
        # Perform the operation
        func(pathname)
        # Now setup the wait loop
        if waitall:
            dirname = pathname
        else:
            dirname, name = os.path.split(pathname)
            dirname = dirname or '.'
        # Check for `pathname` to be removed from the filesystem.
        # The exponential backoff of the timeout amounts to a total
        # of ~1 second after which the deletion is probably an error
        # anyway.
        # Testing on an i7@4.3GHz shows that usually only 1 iteration is
        # required when contention occurs.
        timeout = 0.001
        while timeout < 1.0:
            # Note we are only testing for the existence of the file(s) in
            # the contents of the directory regardless of any security or
            # access rights.  If we have made it this far, we have sufficient
            # permissions to do that much using Python's equivalent of the
            # Windows API FindFirstFile.
            # Other Windows APIs can fail or give incorrect results when
            # dealing with files that are pending deletion.
            L = os.listdir(dirname)
            if not (L if waitall else name in L):
                return
            # Increase the timeout and try again
            time.sleep(timeout)
            timeout *= 2
        warnings.warn('tests may fail, delete still pending for ' + pathname,
                      RuntimeWarning, stacklevel=4)

    def _unlink(filename):
        _waitfor(os.unlink, filename)

    def _rmdir(dirname):
        _waitfor(os.rmdir, dirname)

    def _rmtree(path):
        def _rmtree_inner(path):
            for name in _force_run(path, os.listdir, path):
                fullname = os.path.join(path, name)
                try:
                    mode = os.lstat(fullname).st_mode
                except OSError as exc:
                    print("support.rmtree(): os.lstat(%r) failed with %s"
                          % (fullname, exc),
                          file=sys.__stderr__)
                    mode = 0
                if stat.S_ISDIR(mode):
                    _waitfor(_rmtree_inner, fullname, waitall=True)
                    _force_run(fullname, os.rmdir, fullname)
                else:
                    _force_run(fullname, os.unlink, fullname)
        _waitfor(_rmtree_inner, path, waitall=True)
        _waitfor(lambda p: _force_run(p, os.rmdir, p), path)

    def _longpath(path):
        try:
            import ctypes
        except ImportError:
            # No ctypes means we can't expands paths.
            pass
        else:
            buffer = ctypes.create_unicode_buffer(len(path) * 2)
            length = ctypes.windll.kernel32.GetLongPathNameW(path, buffer,
                                                             len(buffer))
            if length:
                return buffer[:length]
        return path
else:
    _unlink = os.unlink
    _rmdir = os.rmdir

    def _rmtree(path):
        import shutil
        try:
            shutil.rmtree(path)
            return
        except OSError:
            pass

        def _rmtree_inner(path):
            for name in _force_run(path, os.listdir, path):
                fullname = os.path.join(path, name)
                try:
                    mode = os.lstat(fullname).st_mode
                except OSError:
                    mode = 0
                if stat.S_ISDIR(mode):
                    _rmtree_inner(fullname)
                    _force_run(path, os.rmdir, fullname)
                else:
                    _force_run(path, os.unlink, fullname)
        _rmtree_inner(path)
        os.rmdir(path)

    def _longpath(path):
        return path


def rmdir(dirname):
    try:
        _rmdir(dirname)
    except FileNotFoundError:
        pass


def rmtree(path):
    try:
        _rmtree(path)
    except FileNotFoundError:
        pass


def fs_is_case_insensitive(directory):
    """Detects if the file system for the specified directory
    is case-insensitive."""
    import tempfile

    with tempfile.NamedTemporaryFile(dir=directory) as base:
        base_path = base.name
        case_path = base_path.upper()
        if case_path == base_path:
            case_path = base_path.lower()
        try:
            return os.path.samefile(base_path, case_path)
        except FileNotFoundError:
            return False


@contextlib.contextmanager
def change_cwd(path, quiet=False):
    """Return a context manager that changes the current working directory.

    Arguments:

      path: the directory to use as the temporary current working directory.

      quiet: if False (the default), the context manager raises an exception
        on error.  Otherwise, it issues only a warning and keeps the current
        working directory the same.

    """
    saved_dir = os.getcwd()
    try:
        os.chdir(os.path.realpath(path))
    except OSError as exc:
        if not quiet:
            raise
        warnings.warn(
            f"tests may fail, unable to change the current working "
            f"directory to {path!r}: {exc}",
            RuntimeWarning,
            stacklevel=3,
        )
    try:
        yield os.getcwd()
    finally:
        os.chdir(saved_dir)


try:
    import ctypes

    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    ERROR_FILE_NOT_FOUND = 2
    DDD_REMOVE_DEFINITION = 2
    DDD_EXACT_MATCH_ON_REMOVE = 4
    DDD_NO_BROADCAST_SYSTEM = 8
except (ImportError, AttributeError):

    @contextlib.contextmanager
    def subst_drive(path):
        raise pytest.skip("ctypes or kernel32 is not available")

else:

    @contextlib.contextmanager
    def subst_drive(path):
        """Temporarily yield a substitute drive for a given path."""
        for c in reversed(string.ascii_uppercase):
            drive = f"{c}:"
            if (
                not kernel32.QueryDosDeviceW(drive, None, 0)
                and ctypes.get_last_error() == ERROR_FILE_NOT_FOUND
            ):
                break
        else:
            raise pytest.skip("no available logical drive")
        if not kernel32.DefineDosDeviceW(DDD_NO_BROADCAST_SYSTEM, drive, path):
            raise ctypes.WinError(ctypes.get_last_error())
        try:
            yield drive
        finally:
            if not kernel32.DefineDosDeviceW(
                DDD_REMOVE_DEFINITION | DDD_EXACT_MATCH_ON_REMOVE, drive, path
            ):
                raise ctypes.WinError(ctypes.get_last_error())


os_helper = SimpleNamespace(
    rmtree=rmtree,
    can_symlink=can_symlink,
    skip_unless_symlink=skip_unless_symlink,
    TESTFN=TESTFN,
    _longpath=_longpath,
    EnvironmentVarGuard=EnvironmentVarGuard,
    skip_unless_working_chmod=skip_unless_working_chmod,
    fs_is_case_insensitive=fs_is_case_insensitive,
    change_cwd=change_cwd,
    subst_drive=subst_drive,
)



import warnings

import fsspec
import pytest

from upath import UPath
from upath.implementations.cloud import GCSPath

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import extends_base
from ..utils import overrides_base
from ..utils import posixify
from ..utils import skip_on_windows


@skip_on_windows
class TestGCSPath(BaseTests, metaclass=OverrideMeta):
    SUPPORTS_EMPTY_DIRS = False

    @pytest.fixture(autouse=True, scope="function")
    def path(self, gcs_fixture):
        path, endpoint_url = gcs_fixture
        self.path = UPath(path, endpoint_url=endpoint_url, token="anon")

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, GCSPath)

    @extends_base
    def test_rmdir(self):
        dirname = "rmdir_test"
        mock_dir = self.path.joinpath(dirname)
        mock_dir.joinpath("test.txt").write_text("hello")
        mock_dir.fs.invalidate_cache()
        mock_dir.rmdir()
        assert not mock_dir.exists()
        with pytest.raises(NotADirectoryError):
            self.path.joinpath("file1.txt").rmdir()


@skip_on_windows
def test_mkdir_in_empty_bucket(docker_gcs):
    fs = fsspec.filesystem("gcs", endpoint_url=docker_gcs, token="anon")
    fs.mkdir("my-fresh-bucket")
    assert "my-fresh-bucket/" in fs.buckets
    fs.invalidate_cache()
    del fs

    UPath(
        "gs://my-fresh-bucket/some-dir/another-dir/file",
        endpoint_url=docker_gcs,
        token="anon",
    ).parent.mkdir(parents=True, exist_ok=True)


@skip_on_windows
@pytest.mark.xfail(reason="gcsfs returns isdir false")
def test_copy__object_key_collides_with_dir_prefix(docker_gcs, tmp_path):
    gcs = fsspec.filesystem(
        "gcs",
        endpoint_url=docker_gcs,
        token="anon",
        use_listings_cache=False,
    )
    bucket = "copy_into_collision_bucket"
    gcs.mkdir(bucket)
    # gcs.mkdir(bucket + "/src" + "/common_prefix/")
    # object under common prefix as key
    gcs.pipe_file(f"{bucket}/src/common_prefix", b"hello world")
    # store more objects with same prefix
    gcs.pipe_file(f"{bucket}/src/common_prefix/file1.txt", b"1")
    gcs.pipe_file(f"{bucket}/src/common_prefix/file2.txt", b"2")
    gcs.invalidate_cache()

    # make sure the sources have a collision
    assert gcs.isfile(f"{bucket}/src/common_prefix")
    assert gcs.isdir(f"{bucket}/src/common_prefix")  # BROKEN in gcsfs
    assert gcs.isfile(f"{bucket}/src/common_prefix/file1.txt")
    assert gcs.isfile(f"{bucket}/src/common_prefix/file2.txt")
    # prepare source and destination
    src = UPath(f"gs://{bucket}/src", endpoint_url=docker_gcs, token="anon")
    dst = UPath(tmp_path)

    def on_collision_rename_file(src, dst):
        warnings.warn(
            f"{src!s} collides with prefix. Renaming target file object to {dst!s}",
            UserWarning,
            stacklevel=3,
        )
        return (
            dst.with_suffix(dst.suffix + ".COLLISION"),
            dst,
        )

    # perform copy
    src.copy_into(dst, on_name_collision=on_collision_rename_file)

    # check results
    dst_files = sorted(posixify(x.relative_to(tmp_path)) for x in dst.glob("**/*"))
    assert dst_files == [
        "src",
        "src/common_prefix",
        "src/common_prefix.COLLISION",
        "src/common_prefix/file1.txt",
        "src/common_prefix/file2.txt",
    ]



import tarfile

import pytest

from upath import UPath
from upath.implementations.tar import TarPath

from ..cases import JoinablePathTests
from ..cases import NonWritablePathTests
from ..cases import ReadablePathTests
from ..utils import OverrideMeta
from ..utils import overrides_base


@pytest.fixture(scope="function")
def tarred_testdir_file(local_testdir, tmp_path_factory):
    base = tmp_path_factory.mktemp("tarpath")
    tar_path = base / "test.tar"
    with tarfile.TarFile(tar_path, "w") as tf:
        tf.add(local_testdir, arcname="", recursive=True)
    return str(tar_path)


class TestTarPath(
    JoinablePathTests,
    ReadablePathTests,
    NonWritablePathTests,
    metaclass=OverrideMeta,
):

    @pytest.fixture(autouse=True)
    def path(self, tarred_testdir_file):
        self.path = UPath("tar://", fo=tarred_testdir_file)
        # self.prepare_file_system()  done outside of UPath

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, TarPath)


@pytest.fixture(scope="function")
def tarred_testdir_file_in_memory(tarred_testdir_file, clear_fsspec_memory_cache):
    p = UPath(tarred_testdir_file, protocol="file")
    t = p.move(UPath("memory:///mytarfile.tar"))
    assert t.protocol == "memory"
    assert t.exists()
    yield t.as_uri()


class TestChainedTarPath(TestTarPath):

    @pytest.fixture(autouse=True)
    def path(self, tarred_testdir_file_in_memory):
        self.path = UPath("tar://::memory:///mytarfile.tar")



from __future__ import annotations

import sys
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sequence
from typing import IO
from typing import TYPE_CHECKING
from typing import Any
from typing import BinaryIO
from typing import Callable
from typing import Literal
from typing import TextIO
from typing import TypeVar
from typing import overload
from urllib.parse import SplitResult

from fsspec import AbstractFileSystem
from pathlib_abc import vfspath

from upath._chain import Chain
from upath._chain import ChainSegment
from upath.core import UnsupportedOperation
from upath.core import UPath
from upath.types import UNSET_DEFAULT
from upath.types import JoinablePathLike
from upath.types import OnNameCollisionFunc
from upath.types import PathInfo
from upath.types import ReadablePath
from upath.types import ReadablePathLike
from upath.types import StatResultType
from upath.types import SupportsPathLike
from upath.types import UPathParser
from upath.types import WritablePathLike

if TYPE_CHECKING:
    if sys.version_info > (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from pydantic import GetCoreSchemaHandler
    from pydantic_core.core_schema import CoreSchema

__all__ = [
    "ProxyUPath",
]

T = TypeVar("T")


class classmethod_or_method(classmethod):
    """A decorator that can be used as a classmethod or an instance method.

    When called on the class, it behaves like a classmethod.
    When called on an instance, it behaves like an instance method.

    """

    def __get__(
        self,
        instance: T | None,
        owner: type[T] | None = None,
        /,
    ) -> Callable[..., T]:
        if instance is None:
            return self.__func__.__get__(owner)
        else:
            return self.__func__.__get__(instance)


class ProxyUPath:
    """ProxyUPath base class

    ProxyUPath should be used when you want to extend the UPath class
    interface with additional methods, but still want to support
    all supported upath implementations.

    """

    __slots__ = ("__wrapped__",)

    # TODO: think about if and how to handle these
    #  _transform_init_args
    #  _parse_storage_options
    #  _fs_factory
    #  _protocol_dispatch

    # === non-public methods / attributes =============================

    @classmethod
    def _from_upath(cls, upath: UPath, /) -> Self:
        if isinstance(upath, cls):
            return upath  # type: ignore[unreachable]
        else:
            obj = object.__new__(cls)
            obj.__wrapped__ = upath
            return obj

    @property
    def _chain(self):
        try:
            return self.__wrapped__._chain
        except AttributeError:
            return Chain(
                ChainSegment(
                    path=self.__wrapped__.path,
                    protocol=self.__wrapped__.protocol,
                    storage_options=dict(self.__wrapped__.storage_options),
                ),
            )

    # === wrapped interface ===========================================

    def __init__(
        self,
        *args: JoinablePathLike,
        protocol: str | None = None,
        **storage_options: Any,
    ) -> None:
        self.__wrapped__ = UPath(*args, protocol=protocol, **storage_options)

    @property
    def parser(self) -> UPathParser:
        return self.__wrapped__.parser

    def with_segments(self, *pathsegments: JoinablePathLike) -> Self:
        return self._from_upath(self.__wrapped__.with_segments(*pathsegments))

    def __str__(self) -> str:
        return self.__wrapped__.__str__()

    def __vfspath__(self) -> str:
        return self.__wrapped__.__vfspath__()

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({self.__wrapped__.path!r}, protocol={self.protocol!r})"
        )

    @property
    def parts(self) -> Sequence[str]:
        return self.__wrapped__.parts

    def with_name(self, name: str) -> Self:
        return self._from_upath(self.__wrapped__.with_name(name))

    @property
    def info(self) -> PathInfo:
        return self.__wrapped__.info

    def iterdir(self) -> Iterator[Self]:
        for pth in self.__wrapped__.iterdir():
            yield self._from_upath(pth)

    def __open_reader__(self) -> BinaryIO:
        return self.__wrapped__.__open_reader__()

    def readlink(self) -> Self:
        return self._from_upath(self.__wrapped__.readlink())

    def symlink_to(
        self,
        target: ReadablePathLike,
        target_is_directory: bool = False,
    ) -> None:
        if not isinstance(target, str):
            target = vfspath(target)
        self.__wrapped__.symlink_to(target, target_is_directory=target_is_directory)

    def mkdir(
        self,
        mode: int = 0o777,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        self.__wrapped__.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    def __open_writer__(self, mode: Literal["a", "w", "x"]) -> BinaryIO:
        return self.__wrapped__.__open_writer__(mode)

    @overload
    def open(
        self,
        mode: Literal["r", "w", "a"] = "r",
        buffering: int = ...,
        encoding: str = ...,
        errors: str = ...,
        newline: str = ...,
        **fsspec_kwargs: Any,
    ) -> TextIO: ...

    @overload
    def open(
        self,
        mode: Literal["rb", "wb", "ab"],
        buffering: int = ...,
        encoding: str = ...,
        errors: str = ...,
        newline: str = ...,
        **fsspec_kwargs: Any,
    ) -> BinaryIO: ...

    @overload
    def open(
        self,
        mode: str,
        *args: Any,
        **fsspec_kwargs: Any,
    ) -> IO[Any]: ...

    def open(
        self,
        mode: str = "r",
        buffering: int = UNSET_DEFAULT,
        encoding: str | None = UNSET_DEFAULT,
        errors: str | None = UNSET_DEFAULT,
        newline: str | None = UNSET_DEFAULT,
        **fsspec_kwargs: Any,
    ) -> IO[Any]:
        return self.__wrapped__.open(
            mode,
            buffering,
            encoding,
            errors,
            newline,
            **fsspec_kwargs,
        )

    def stat(
        self,
        *,
        follow_symlinks=True,
    ) -> StatResultType:
        return self.__wrapped__.stat(follow_symlinks=follow_symlinks)

    def lstat(self) -> StatResultType:
        return self.__wrapped__.stat(follow_symlinks=False)

    def chmod(self, mode: int, *, follow_symlinks: bool = True) -> None:
        self.__wrapped__.chmod(mode=mode, follow_symlinks=follow_symlinks)

    def exists(self, *, follow_symlinks=True) -> bool:
        return self.__wrapped__.exists(follow_symlinks=follow_symlinks)

    def is_dir(self) -> bool:
        return self.__wrapped__.is_dir()

    def is_file(self) -> bool:
        return self.__wrapped__.is_file()

    def is_mount(self) -> bool:
        return self.__wrapped__.is_mount()

    def is_symlink(self) -> bool:
        return self.__wrapped__.is_symlink()

    def is_junction(self) -> bool:
        return self.__wrapped__.is_junction()

    def is_block_device(self) -> bool:
        return self.__wrapped__.is_block_device()

    def is_char_device(self) -> bool:
        return self.__wrapped__.is_char_device()

    def is_fifo(self) -> bool:
        return self.__wrapped__.is_fifo()

    def is_socket(self) -> bool:
        return self.__wrapped__.is_socket()

    def is_reserved(self) -> bool:
        return self.__wrapped__.is_reserved()

    def expanduser(self) -> Self:
        return self._from_upath(self.__wrapped__.expanduser())

    def glob(
        self,
        pattern: str,
        *,
        case_sensitive: bool | None = None,
        recurse_symlinks: bool = False,
    ) -> Iterator[Self]:
        for p in self.__wrapped__.glob(
            pattern, case_sensitive=case_sensitive, recurse_symlinks=recurse_symlinks
        ):
            yield self._from_upath(p)

    def rglob(
        self,
        pattern: str,
        *,
        case_sensitive: bool | None = None,
        recurse_symlinks: bool = False,
    ) -> Iterator[Self]:
        for p in self.__wrapped__.rglob(
            pattern, case_sensitive=case_sensitive, recurse_symlinks=recurse_symlinks
        ):
            yield self._from_upath(p)

    def owner(self) -> str:
        return self.__wrapped__.owner()

    def group(self) -> str:
        return self.__wrapped__.group()

    def absolute(self) -> Self:
        return self._from_upath(self.__wrapped__.absolute())

    def is_absolute(self) -> bool:
        return self.__wrapped__.is_absolute()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__wrapped__.__eq__(other.__wrapped__)

    def __hash__(self) -> int:
        return self.__wrapped__.__hash__()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__wrapped__.__ne__(other.__wrapped__)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__wrapped__.__lt__(other.__wrapped__)

    def __le__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__wrapped__.__le__(other.__wrapped__)

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__wrapped__.__gt__(other.__wrapped__)

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__wrapped__.__ge__(other.__wrapped__)

    def resolve(self, strict: bool = False) -> Self:
        return self._from_upath(self.__wrapped__.resolve(strict=strict))

    def touch(self, mode=0o666, exist_ok=True) -> None:
        self.__wrapped__.touch(mode=mode, exist_ok=exist_ok)

    def lchmod(self, mode: int) -> None:
        self.__wrapped__.lchmod(mode=mode)

    def unlink(self, missing_ok: bool = False) -> None:
        self.__wrapped__.unlink(missing_ok=missing_ok)

    def rmdir(self, recursive: bool = UNSET_DEFAULT) -> None:  # fixme: non-standard
        kwargs: dict[str, Any] = {}
        if recursive is not UNSET_DEFAULT:
            kwargs["recursive"] = recursive
        self.__wrapped__.rmdir(**kwargs)

    def rename(
        self,
        target: WritablePathLike,
        *,  # note: non-standard compared to pathlib
        recursive: bool = UNSET_DEFAULT,
        maxdepth: int | None = UNSET_DEFAULT,
        **kwargs: Any,
    ) -> Self:
        if recursive is not UNSET_DEFAULT:
            kwargs["recursive"] = recursive
        if maxdepth is not UNSET_DEFAULT:
            kwargs["maxdepth"] = maxdepth
        return self._from_upath(
            self.__wrapped__.rename(
                target.__wrapped__ if isinstance(target, ProxyUPath) else target,
                **kwargs,
            )
        )

    def replace(self, target: WritablePathLike) -> Self:
        return self._from_upath(self.__wrapped__.replace(target))

    @property
    def drive(self) -> str:
        return self.__wrapped__.drive

    @property
    def root(self) -> str:
        return self.__wrapped__.root

    def __reduce__(self):
        return type(self)._from_upath, (self.__wrapped__,)

    @classmethod
    def from_uri(cls, uri: str, **storage_options: Any) -> Self:
        return cls(uri, **storage_options)

    def as_uri(self) -> str:
        return self.__wrapped__.as_uri()

    def as_posix(self) -> str:
        return self.__wrapped__.as_posix()

    def samefile(self, other_path) -> bool:
        return self.__wrapped__.samefile(other_path)

    @classmethod_or_method
    def cwd(cls_or_self) -> Self:  # noqa: B902
        if isinstance(cls_or_self, type):
            raise UnsupportedOperation(".cwd() not supported")
        else:
            return cls_or_self._from_upath(cls_or_self.__wrapped__.cwd())

    @classmethod
    def home(cls) -> Self:
        raise UnsupportedOperation(".home() not supported")

    def relative_to(  # type: ignore[override]
        self,
        other,
        /,
        *_deprecated,
        walk_up=False,
    ) -> Self:
        if isinstance(other, ProxyUPath):
            other = other.__wrapped__
        return self._from_upath(
            self.__wrapped__.relative_to(other, *_deprecated, walk_up=walk_up)
        )

    def is_relative_to(self, other, /, *_deprecated) -> bool:  # type: ignore[override]
        if not isinstance(other, str):
            other = vfspath(other)
        return self.__wrapped__.is_relative_to(other, *_deprecated)

    def hardlink_to(self, target: ReadablePathLike) -> None:
        if not isinstance(target, str):
            target = vfspath(target)
        return self.__wrapped__.hardlink_to(target)

    def match(self, pattern: str, *, case_sensitive: bool | None = None) -> bool:
        return self.__wrapped__.match(pattern)

    @property
    def protocol(self) -> str:
        return self.__wrapped__.protocol

    @property
    def storage_options(self) -> Mapping[str, Any]:
        return self.__wrapped__.storage_options

    @property
    def fs(self) -> AbstractFileSystem:
        return self.__wrapped__.fs

    @property
    def path(self) -> str:
        return self.__wrapped__.path

    def joinuri(self, uri: JoinablePathLike) -> Self:
        return self._from_upath(self.__wrapped__.joinuri(uri))

    @property
    def _url(self) -> SplitResult:
        return self.__wrapped__._url

    def read_bytes(self) -> bytes:
        return self.__wrapped__.read_bytes()

    def read_text(
        self,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> str:
        return self.__wrapped__.read_text(
            encoding=encoding, errors=errors, newline=newline
        )

    def walk(
        self,
        top_down: bool = True,
        on_error: Callable[[Exception], Any] | None = None,
        follow_symlinks: bool = False,
    ) -> Iterator[tuple[Self, list[str], list[str]]]:
        for pth, dirnames, filenames in self.__wrapped__.walk(
            top_down=top_down, on_error=on_error, follow_symlinks=follow_symlinks
        ):
            yield self._from_upath(pth), dirnames, filenames

    def copy(self, target: WritablePathLike, **kwargs: Any) -> Self:  # type: ignore[override]  # noqa: E501
        return self._from_upath(self.__wrapped__.copy(target, **kwargs))

    def copy_into(self, target_dir: WritablePathLike, **kwargs: Any) -> Self:  # type: ignore[override]  # noqa: E501
        return self._from_upath(self.__wrapped__.copy_into(target_dir, **kwargs))

    def move(self, target: WritablePathLike, **kwargs: Any) -> Self:  # type: ignore[override]  # noqa: E501
        return self._from_upath(self.__wrapped__.move(target, **kwargs))

    def move_into(self, target_dir: WritablePathLike, **kwargs: Any) -> Self:  # type: ignore[override]  # noqa: E501
        return self._from_upath(self.__wrapped__.move_into(target_dir, **kwargs))

    def write_bytes(self, data: bytes) -> int:
        return self.__wrapped__.write_bytes(data)

    def write_text(
        self,
        data: str,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> int:
        return self.__wrapped__.write_text(
            data, encoding=encoding, errors=errors, newline=newline
        )

    def _copy_from(
        self,
        source: ReadablePath | Self,
        follow_symlinks: bool = True,
        on_name_collision: OnNameCollisionFunc | None = None,
        **kwargs: Any,
    ) -> None:
        self.__wrapped__._copy_from(source, follow_symlinks=follow_symlinks, on_name_collision=on_name_collision, **kwargs)  # type: ignore  # noqa: E501

    @property
    def anchor(self) -> str:
        return self.__wrapped__.anchor

    @property
    def name(self) -> str:
        return self.__wrapped__.name

    @property
    def suffix(self) -> str:
        return self.__wrapped__.suffix

    @property
    def suffixes(self) -> list[str]:
        return self.__wrapped__.suffixes

    @property
    def stem(self) -> str:
        return self.__wrapped__.stem

    def with_stem(self, stem: str) -> Self:
        return self._from_upath(self.__wrapped__.with_stem(stem))

    def with_suffix(self, suffix: str) -> Self:
        return self._from_upath(self.__wrapped__.with_suffix(suffix))

    def joinpath(self, *pathsegments: JoinablePathLike) -> Self:
        return self._from_upath(self.__wrapped__.joinpath(*pathsegments))

    def __truediv__(self, other: str | Self) -> Self:
        return self._from_upath(
            self.__wrapped__.__truediv__(other)  # type: ignore[operator]
        )

    def __rtruediv__(self, other: str | Self) -> Self:
        return self._from_upath(
            self.__wrapped__.__rtruediv__(other)  # type: ignore[operator]
        )

    @property
    def parent(self) -> Self:
        return self._from_upath(self.__wrapped__.parent)

    @property
    def parents(self) -> Sequence[Self]:
        return tuple(self._from_upath(p) for p in self.__wrapped__.parents)

    def full_match(
        self,
        pattern: str | SupportsPathLike,
        *,
        case_sensitive: bool | None = None,
    ) -> bool:
        return self.__wrapped__.full_match(pattern, case_sensitive=case_sensitive)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        cs = UPath.__get_pydantic_core_schema__.__func__  # type: ignore[attr-defined]
        return cs(cls, source_type, handler)


UPath.register(ProxyUPath)



import os
import sys
from contextlib import nullcontext
from pathlib import PosixPath
from pathlib import WindowsPath

import pytest

from upath import UnsupportedOperation
from upath import UPath
from upath.implementations.local import PosixUPath
from upath.implementations.local import WindowsUPath


def test_provides_unsupportedoperation():
    assert issubclass(UnsupportedOperation, NotImplementedError)


_win_only = pytest.mark.skipif(os.name != "nt", reason="windows only")
_posix_only = pytest.mark.skipif(os.name == "nt", reason="posix only")


@pytest.mark.parametrize(
    "wrong_cls",
    [
        pytest.param(PosixPath, marks=_win_only),
        pytest.param(PosixUPath, marks=_win_only),
        pytest.param(WindowsPath, marks=_posix_only),
        pytest.param(WindowsUPath, marks=_posix_only),
    ],
)
def test_unsupportedoperation_catches_pathlib_errors(wrong_cls):
    with pytest.raises(UnsupportedOperation):
        wrong_cls(".")


@pytest.fixture(params=["", "file", "memory"])
def pth(request, tmp_path, clear_fsspec_memory_cache):
    pth = UPath(tmp_path, protocol=request.param)
    yield pth


@pytest.fixture
def pth_file(pth):
    f = pth.joinpath("testfile")
    f.touch()
    return f


@pytest.mark.parametrize(
    "pth",
    [
        pytest.param(
            "",
            marks=pytest.mark.xfail(
                sys.version_info < (3, 10), reason="parents getitem differs"
            ),
        ),
        "file",
        "memory",
    ],
    indirect=True,
)
def test_signature_parents_getitem(pth):
    pth.parents[-1]
    pth.parents[0:2]


def test_signature_stat_follow_symlinks(pth_file):
    pth_file.stat(follow_symlinks=True)


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=r".*write_text\(\) newline is currently ignored",
                )
                if sys.version_info < (3, 10)
                else nullcontext()
            ),
        ),
        ("file", nullcontext()),
        ("memory", nullcontext()),
    ],
    indirect=["pth"],
)
def test_signature_write_text_newline(pth, ctx):
    with ctx:
        _ = pth.joinpath("somename").write_text("test", newline="\n")


@pytest.mark.parametrize("pth", ["", "file"], indirect=True)
def test_signature_chmod_follow_symlinks(pth):
    _ = pth.chmod(0o777, follow_symlinks=True)


def test_signature_match_pathlike(pth):
    _ = pth.match(UPath("*.py"))


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=r".*match\(\): case_sensitive is currently ignored",
                )
                if sys.version_info < (3, 12)
                else nullcontext()
            ),
        ),
        (
            "file",
            pytest.warns(
                UserWarning, match=r".*match\(\): case_sensitive is currently ignored"
            ),
        ),
        (
            "memory",
            pytest.warns(
                UserWarning, match=r".*match\(\): case_sensitive is currently ignored"
            ),
        ),
    ],
    indirect=["pth"],
)
def test_signature_match_case_sensitive(pth, ctx):
    with ctx:
        _ = pth.match("*.py", case_sensitive=True)


def test_signature_exists_follow_symlinks(pth):
    _ = pth.exists(follow_symlinks=True)


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=r".*glob\(\): case_sensitive is currently ignored",
                )
                if sys.version_info < (3, 12)
                else nullcontext()
            ),
        ),
        (
            "file",
            pytest.warns(
                UserWarning, match=r".*glob\(\): case_sensitive is currently ignored"
            ),
        ),
        (
            "memory",
            pytest.warns(
                UserWarning, match=r".*glob\(\): case_sensitive is currently ignored"
            ),
        ),
    ],
    indirect=["pth"],
)
def test_signature_glob_case_sensitive(pth, ctx):
    with ctx:
        _ = list(pth.parent.glob("*", case_sensitive=True))


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=r".*(glob|rglob)\(\): case_sensitive is currently ignored",
                )
                if sys.version_info < (3, 12)
                else nullcontext()
            ),
        ),
        (
            "file",
            pytest.warns(
                UserWarning,
                match=r".*(glob|rglob)\(\): case_sensitive is currently ignored",
            ),
        ),
        (
            "memory",
            pytest.warns(
                UserWarning,
                match=r".*(glob|rglob)\(\): case_sensitive is currently ignored",
            ),
        ),
    ],
    indirect=["pth"],
)
def test_signature_rglob_case_sensitive(pth, ctx):
    with ctx:
        _ = list(pth.parent.rglob("*", case_sensitive=True))


def test_signature_relative_to_walk_up(pth):
    _ = pth.relative_to(pth.parent.parent, walk_up=False)


def test_signature_is_file_follow_symlinks(pth):
    _ = pth.is_file(follow_symlinks=True)


def test_signature_is_dir_follow_symlinks(pth):
    _ = pth.is_dir(follow_symlinks=True)


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=r".*read_text\(\): newline is currently ignored",
                )
                if sys.version_info < (3, 13)
                else nullcontext()
            ),
        ),
        ("file", nullcontext()),
        ("memory", nullcontext()),
    ],
    indirect=["pth"],
)
def test_signature_read_text_newline(pth, ctx):
    p0 = pth.joinpath("test")
    p0.write_text("test")
    with ctx:
        _ = p0.read_text(newline="\n")


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=r".*glob\(\): recurse_symlinks=True is currently ignored",
                )
                if sys.version_info < (3, 13)
                else nullcontext()
            ),
        ),
        (
            "file",
            pytest.warns(
                UserWarning,
                match=r".*glob\(\): recurse_symlinks=True is currently ignored",
            ),
        ),
        (
            "memory",
            pytest.warns(
                UserWarning,
                match=r".*glob\(\): recurse_symlinks=True is currently ignored",
            ),
        ),
    ],
    indirect=["pth"],
)
def test_signature_glob_recurse_symlinks(pth, ctx):
    with ctx:
        _ = list(pth.parent.glob("**/*.py", recurse_symlinks=True))


def test_signature_glob_pathlike(pth):
    _ = list(pth.parent.glob(UPath("**/*.py")))


@pytest.mark.parametrize(
    "pth,ctx",
    [
        (
            "",
            (
                pytest.warns(
                    UserWarning,
                    match=(
                        r".*(glob|rglob)\(\):"
                        r" recurse_symlinks=True is currently ignored"
                    ),
                )
                if sys.version_info < (3, 13)
                else nullcontext()
            ),
        ),
        (
            "file",
            pytest.warns(
                UserWarning,
                match=r".*(glob|rglob)\(\): recurse_symlinks=True is currently ignored",
            ),
        ),
        (
            "memory",
            pytest.warns(
                UserWarning,
                match=r".*(glob|rglob)\(\): recurse_symlinks=True is currently ignored",
            ),
        ),
    ],
    indirect=["pth"],
)
def test_signature_rglob_recurse_symlinks(pth, ctx):
    with ctx:
        _ = list(pth.parent.rglob("*.py", recurse_symlinks=True))


def test_signature_rglob_pathlike(pth):
    _ = list(pth.parent.rglob(UPath("*.py")))


@pytest.mark.parametrize("pth", [""], indirect=True)
@pytest.mark.skipif(os.name == "nt", reason="owner/group not available on windows")
def test_signature_owner_follow_symlinks(pth):
    _ = pth.owner(follow_symlinks=True)


@pytest.mark.parametrize("pth", [""], indirect=True)
@pytest.mark.skipif(os.name == "nt", reason="owner/group not available on windows")
def test_signature_group_follow_symlinks(pth):
    _ = pth.group(follow_symlinks=True)



"""pathlib_abc exports for compatibility with pathlib."""

import sys
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import BinaryIO
from typing import Callable
from typing import Iterator
from typing import Literal
from typing import Protocol
from typing import Sequence
from typing import TextIO
from typing import TypeVar
from typing import runtime_checkable

if sys.version_info > (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

class JoinablePath(ABC):
    __slots__ = ()

    @property
    @abstractmethod
    def parser(self) -> PathParser: ...
    @abstractmethod
    def with_segments(self, *pathsegments: str) -> Self: ...
    @abstractmethod
    def __vfspath__(self) -> str: ...
    @property
    def anchor(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def suffix(self) -> str: ...
    @property
    def suffixes(self) -> list[str]: ...
    @property
    def stem(self) -> str: ...
    def with_name(self, name: str) -> Self: ...
    def with_stem(self, stem: str) -> Self: ...
    def with_suffix(self, suffix: str) -> Self: ...
    @property
    def parts(self) -> Sequence[str]: ...
    def joinpath(self, *pathsegments: str) -> Self: ...
    def __truediv__(self, key: str) -> Self: ...
    def __rtruediv__(self, key: str) -> Self: ...
    @property
    def parent(self) -> Self: ...
    @property
    def parents(self) -> Sequence[Self]: ...
    def full_match(self, pattern: str) -> bool: ...

OnErrorCallable = Callable[[Exception], Any]
T = TypeVar("T", bound="WritablePath")

class ReadablePath(JoinablePath):
    __slots__ = ()

    @property
    @abstractmethod
    def info(self) -> PathInfo: ...
    @abstractmethod
    def __open_reader__(self) -> BinaryIO: ...
    def read_bytes(self) -> bytes: ...
    def read_text(
        self,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
    ) -> str: ...
    @abstractmethod
    def iterdir(self) -> Iterator[Self]: ...
    def glob(self, pattern: str, *, recurse_symlinks: bool = ...) -> Iterator[Self]: ...
    def walk(
        self,
        top_down: bool = ...,
        on_error: OnErrorCallable | None = ...,
        follow_symlinks: bool = ...,
    ) -> Iterator[tuple[Self, list[str], list[str]]]: ...
    @abstractmethod
    def readlink(self) -> Self: ...
    def copy(self, target: T, **kwargs: Any) -> T: ...
    def copy_into(self, target_dir: T, **kwargs: Any) -> T: ...

class WritablePath(JoinablePath):
    __slots__ = ()

    @abstractmethod
    def symlink_to(
        self, target: ReadablePath, target_is_directory: bool = ...
    ) -> None: ...
    @abstractmethod
    def mkdir(self) -> None: ...
    @abstractmethod
    def __open_writer__(self, mode: Literal["a", "w", "x"]) -> BinaryIO: ...
    def write_bytes(self, data: bytes) -> int: ...
    def write_text(
        self,
        data: str,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
    ) -> int: ...
    def _copy_from(self, source: ReadablePath, follow_symlinks: bool = ...) -> None: ...

@runtime_checkable
class PathParser(Protocol):
    sep: str
    altsep: str | None

    def split(self, path: str) -> tuple[str, str]: ...
    def splitext(self, path: str) -> tuple[str, str]: ...
    def normcase(self, path: str) -> str: ...

@runtime_checkable
class PathInfo(Protocol):
    def exists(self, *, follow_symlinks: bool = True) -> bool: ...
    def is_dir(self, *, follow_symlinks: bool = True) -> bool: ...
    def is_file(self, *, follow_symlinks: bool = True) -> bool: ...
    def is_symlink(self) -> bool: ...

class SupportsOpenReader(Protocol):
    def __open_reader__(self) -> BinaryIO: ...

class SupportsOpenWriter(Protocol):
    def __open_writer__(self, mode: Literal["a", "w", "x"]) -> BinaryIO: ...

class SupportsOpenUpdater(Protocol):
    def __open_updater__(self, mode: Literal["r+", "w+", "+r", "+w"]) -> BinaryIO: ...

def vfsopen(
    obj: SupportsOpenReader | SupportsOpenWriter | SupportsOpenUpdater,
    mode="r",
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
) -> BinaryIO | TextIO: ...

class SupportsVFSPath(Protocol):
    def __vfspath__(self) -> str: ...

def vfspath(obj: SupportsVFSPath) -> str: ...



import os
import sys
from contextlib import nullcontext

import pytest

from upath import UnsupportedOperation
from upath import UPath
from upath.extensions import ProxyUPath
from upath.implementations.local import FilePath
from upath.implementations.local import PosixUPath
from upath.implementations.local import WindowsUPath
from upath.implementations.memory import MemoryPath

from .cases import BaseTests
from .utils import OverrideMeta
from .utils import extends_base
from .utils import overrides_base


class TestProxyMemoryPath(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        if not local_testdir.startswith("/"):
            local_testdir = "/" + local_testdir
        self.path = ProxyUPath(f"memory:{local_testdir}")
        self.prepare_file_system()

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, ProxyUPath)

    @extends_base
    def test_is_not_wrapped_class(self):
        assert not isinstance(self.path, MemoryPath)


class TestProxyFilePath(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        self.path = ProxyUPath(f"file://{local_testdir}")
        self.prepare_file_system()

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, ProxyUPath)

    @extends_base
    def test_is_not_wrapped_class(self):
        assert not isinstance(self.path, FilePath)

    @overrides_base
    def test_chmod(self):
        self.path.joinpath("file1.txt").chmod(777)

    @overrides_base
    def test_cwd(self):
        # ProxyUPath.cwd() works differently on the instance
        self.path.cwd()
        with pytest.raises(UnsupportedOperation):
            type(self.path).cwd()


class TestProxyPathlibPath(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        self.path = ProxyUPath(f"{local_testdir}")
        self.prepare_file_system()

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, ProxyUPath)

    @extends_base
    def test_is_not_wrapped_class(self):
        assert not isinstance(self.path, (PosixUPath, WindowsUPath))

    @overrides_base
    def test_chmod(self):
        self.path.joinpath("file1.txt").chmod(777)

    @overrides_base
    @pytest.mark.skipif(
        sys.version_info < (3, 12), reason="storage options only handled in 3.12+"
    )
    def test_eq(self):
        super().test_eq()

    if sys.version_info < (3, 12):

        @overrides_base
        def test_storage_options_dont_affect_hash(self):
            # On Python < 3.12, storage_options trigger warnings for LocalPath
            with pytest.warns(
                UserWarning,
                match=r".*on python <= \(3, 11\) ignores protocol and storage_options",
            ):
                super().test_storage_options_dont_affect_hash()

    @overrides_base
    def test_group(self):
        pytest.importorskip("grp")
        self.path.group()

    @overrides_base
    def test_owner(self):
        pytest.importorskip("pwd")
        self.path.owner()

    @overrides_base
    def test_readlink(self):
        try:
            os.readlink
        except AttributeError:
            pytest.skip("os.readlink not available on this platform")
        with pytest.raises((OSError, UnsupportedOperation)):
            self.path.readlink()

    @overrides_base
    def test_protocol(self):
        assert self.path.protocol == ""

    @overrides_base
    def test_as_uri(self):
        assert self.path.as_uri().startswith("file://")

    if sys.version_info < (3, 10):

        @overrides_base
        def test_lstat(self):
            # On Python < 3.10, stat(follow_symlinks=False) triggers warnings
            with pytest.warns(
                UserWarning,
                match=r".*stat\(\) follow_symlinks=False is currently ignored",
            ):
                st = self.path.lstat()
            assert st is not None

    else:

        @overrides_base
        def test_lstat(self):
            st = self.path.lstat()
            assert st is not None

    @overrides_base
    def test_relative_to(self):
        base = self.path
        child = self.path / "folder1" / "file1.txt"
        relative = child.relative_to(base)
        assert str(relative) == f"folder1{os.sep}file1.txt"

    @overrides_base
    def test_cwd(self):
        self.path.cwd()
        with pytest.raises(UnsupportedOperation):
            type(self.path).cwd()

    @overrides_base
    def test_lchmod(self):
        # setup
        a = self.path.joinpath("a")
        b = self.path.joinpath("b")
        a.touch()
        b.symlink_to(a)

        # see: https://github.com/python/cpython/issues/108660#issuecomment-1854645898
        if hasattr(os, "lchmod") or os.chmod in os.supports_follow_symlinks:
            cm = nullcontext()
        else:
            cm = pytest.raises((UnsupportedOperation, NotImplementedError))
        with cm:
            b.lchmod(mode=0o777)

    @overrides_base
    def test_symlink_to(self):
        self.path.joinpath("link").symlink_to(self.path)

    @overrides_base
    def test_hardlink_to(self):
        try:
            self.path.joinpath("link").hardlink_to(self.path)
        except PermissionError:
            pass  # hardlink may require elevated permissions


def test_custom_subclass():

    class ReversePath(ProxyUPath):
        def read_bytes_reversed(self):
            return self.read_bytes()[::-1]

        def write_bytes_reversed(self, value):
            self.write_bytes(value[::-1])

    b = MemoryPath("memory://base")

    p = b.joinpath("file1")
    p.write_bytes(b"dlrow olleh")

    r = ReversePath("memory://base/file1")
    assert r.read_bytes_reversed() == b"hello world"

    r.parent.joinpath("file2").write_bytes_reversed(b"dlrow olleh")
    assert b.joinpath("file2").read_bytes() == b"hello world"


def test_protocol_dispatch_deprecation_warning():

    class MyPath(UPath):
        _protocol_dispatch = False

    with pytest.warns(DeprecationWarning, match="_protocol_dispatch = False"):
        a = MyPath(".", protocol="memory")

    assert isinstance(a, MyPath)


# Protocol to sample URI mapping for compatibility tests
_PROTOCOL_URIS = {
    "s3": "s3://bucket/folder",
    "gs": "gs://bucket/folder",
    "memory": "memory://folder",
    "file": "file:/tmp/folder",
    "http": "http://example.com/path",
    "": "/tmp/folder",
}

# Generate incompatible combinations
_PROXY_INCOMPATIBLE_CASES = [
    (_PROTOCOL_URIS[uri_protocol], target_protocol)
    for target_protocol in _PROTOCOL_URIS
    for uri_protocol in _PROTOCOL_URIS
    if target_protocol != uri_protocol and uri_protocol != ""
]


@pytest.mark.parametrize("uri,protocol", _PROXY_INCOMPATIBLE_CASES)
def test_proxy_subclass_incompatible_protocol_uri(uri, protocol):
    """Test that ProxyUPath subclasses raise TypeError for incompatible protocols."""

    class MyProxyPath(ProxyUPath):
        pass

    # ProxyUPath wraps the underlying path, so it should also raise TypeError
    with pytest.raises(TypeError, match=r".*incompatible with"):
        MyProxyPath(uri, protocol=protocol)


def test_proxy_upath_copy_from_local(tmp_path):
    """Test ProxyUPath accepts extra kwargs in _copy_from()

    Regression test for https://github.com/fsspec/universal_pathlib/issues/546
    """

    class MyProxyPath(ProxyUPath):
        pass

    source = UPath(tmp_path / "test.txt")
    source.write_text("hello")

    destination = MyProxyPath("memory://bla/test_copy_from.txt")
    source.move(destination)

    assert destination.read_text() == "hello"



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from upath.core import UPath
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import MemoryStorageOptions

__all__ = ["MemoryPath"]


class MemoryPath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["memory"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[MemoryStorageOptions],
        ) -> None: ...

    @property
    def path(self) -> str:
        path = super().path
        return "/" if path in {"", "."} else path

    def is_absolute(self) -> bool:
        if self._relative_base is None and self.__vfspath__() == "/":
            return True
        return super().is_absolute()

    def __str__(self) -> str:
        s = super().__str__()
        if s.startswith("memory:///"):
            s = s.replace("memory:///", "memory://", 1)
        return s



from __future__ import annotations

import sys
from ftplib import error_perm as FTPPermanentError  # nosec B402
from typing import TYPE_CHECKING

from upath.core import UPath
from upath.types import UNSET_DEFAULT
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from typing import Any
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Self
        from typing import Unpack
    else:
        from typing_extensions import Self
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types import WritablePathLike
    from upath.types.storage_options import FTPStorageOptions

__all__ = ["FTPPath"]


class FTPPath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["ftp"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[FTPStorageOptions],
        ) -> None: ...

    def mkdir(
        self,
        mode: int = 0o777,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        try:
            return super().mkdir(mode, parents, exist_ok)
        except FTPPermanentError as e:
            if e.args[0].startswith("550") and exist_ok:
                return
            raise FileExistsError(str(self)) from e

    def rename(
        self,
        target: WritablePathLike,
        *,  # note: non-standard compared to pathlib
        recursive: bool = UNSET_DEFAULT,
        maxdepth: int | None = UNSET_DEFAULT,
        **kwargs: Any,
    ) -> Self:
        t = super().rename(target, recursive=recursive, maxdepth=maxdepth, **kwargs)
        self_dir = self.parent.path
        t.fs.invalidate_cache(self_dir)
        self.fs.invalidate_cache(self_dir)
        return t



"""see upath/tests/conftest.py for fixtures"""

import pytest  # noqa: F401

from upath import UPath
from upath.implementations.hdfs import HDFSPath

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import overrides_base


@pytest.mark.hdfs
class TestUPathHDFS(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir, hdfs):
        host, user, port = hdfs
        path = f"hdfs:{local_testdir}"
        self.path = UPath(path, host=host, user=user, port=port)

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, HDFSPath)



import pytest
from fsspec import get_filesystem_class

from upath import UnsupportedOperation
from upath import UPath
from upath.implementations.cloud import HfPath

from ..cases import JoinablePathTests
from ..cases import NonWritablePathTests
from ..cases import ReadablePathTests
from ..utils import OverrideMeta
from ..utils import overrides_base

try:
    get_filesystem_class("hf")
except ImportError:
    pytestmark = pytest.mark.skip


def test_hfpath():
    path = UPath("hf://HuggingFaceTB/SmolLM2-135M")
    assert isinstance(path, HfPath)
    try:
        assert path.exists()
    except AssertionError:
        from httpx import ConnectError
        from huggingface_hub import HfApi

        try:
            HfApi().repo_info("HuggingFaceTB/SmolLM2-135M")
        except ConnectError:
            pytest.xfail("No internet connection")
        except Exception as err:
            if "Service Unavailable" in str(err):
                pytest.xfail("HuggingFace API not reachable")
            raise


class TestUPathHf(
    JoinablePathTests,
    ReadablePathTests,
    NonWritablePathTests,
    metaclass=OverrideMeta,
):
    @pytest.fixture(autouse=True, scope="function")
    def path(self, hf_fixture_with_readonly_mocked_hf_api):
        self.path = UPath(hf_fixture_with_readonly_mocked_hf_api)

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, HfPath)

    @overrides_base
    def test_iterdir_parent_iteration(self):
        # HfPath does not support listing all available Repositories
        with pytest.raises(UnsupportedOperation):
            super().test_iterdir_parent_iteration()



"""Tests for relative path functionality."""

import os
import pickle
import re
import tempfile
import warnings
from pathlib import Path

import pytest

from upath import UPath


@pytest.mark.parametrize(
    "protocol,storage_options,path,base",
    [
        ("memory", {}, "memory:///foo/bar/baz.txt", "memory:///foo"),
        ("s3", {"anon": True}, "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo"),
        ("gcs", {"token": "anon"}, "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo"),
        ("http", {"something": 1}, "http://host/foo/bar/baz.txt", "http://host/foo"),
        (
            "https",
            {},
            "https://host/foo/bar/baz.txt",
            "https://host/",
        ),
    ],
)
def test_protocol_storage_options_fs_preserved(protocol, storage_options, path, base):
    """Test that protocol and storage_options are preserved in relative paths."""
    p = UPath(path, protocol=protocol, **storage_options)
    root = UPath(base, protocol=protocol, **storage_options)
    rel = p.relative_to(root)

    assert rel.protocol == protocol
    assert dict(**rel.storage_options) == storage_options
    assert isinstance(rel.fs, type(p.fs))


@pytest.mark.parametrize(
    "protocol,path,base",
    [
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo"),
        ("ftp", "ftp://user:pass@host/foo/bar/baz.txt", "ftp://user:pass@host/foo"),
        ("http", "http://host/foo/bar/baz.txt", "http://host/foo"),
        ("https", "https://host/foo/bar/baz.txt", "https://host/foo"),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo"),
    ],
)
def test_relative_urlpath_raises_without_cwd(protocol, path, base):
    rel = UPath(path, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    with pytest.raises(
        NotImplementedError,
        match=re.escape(f"{type(rel).__name__}.cwd() is unsupported"),
    ):
        rel.cwd()
    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            f"fsspec paths can not be relative and"
            f" {type(rel).__name__}.cwd() is unsupported"
        ),
    ):
        _ = rel.path


@pytest.mark.parametrize(
    "pth,base,rel",
    [
        ("/foo/bar/baz.txt", "/foo", "bar/baz.txt"),
        ("/foo/bar/baz/qux.txt", "/foo/bar", "baz/qux.txt"),
        ("/foo/bar/baz/qux.txt", "/foo/bar/baz", "qux.txt"),
        ("/foo/bar/baz", "/foo/bar/baz", "."),
    ],
)
@pytest.mark.parametrize(
    "protocol",
    [
        "memory",
        "file",
        "",
    ],
)
def test_basic_relative_path_creation(protocol, pth, base, rel):
    rel_pth = UPath(pth, protocol=protocol).relative_to(UPath(base, protocol=protocol))

    assert not rel_pth.is_absolute()
    assert rel_pth.as_posix() == rel


def test_relative_path_validation():
    """Test validation of relative_to arguments."""
    p = UPath("memory:///foo/bar")

    # Different protocols should fail
    with pytest.raises(ValueError, match="incompatible protocols"):
        p.relative_to(UPath("s3://bucket"))

    # Different storage options should fail
    with pytest.raises(ValueError, match="incompatible storage_options"):
        UPath("s3://bucket/file", anon=True).relative_to(
            UPath("s3://bucket", anon=False)
        )


def test_path_not_in_subpath():
    """Test relative_to with paths that don't have a parent-child relationship."""
    p = UPath("memory:///foo/bar")
    other = UPath("memory:///baz")

    with pytest.raises(ValueError, match="is not in the subpath of"):
        p.relative_to(other)


def test_filesystem_operations_fail_without_cwd():
    """Test that filesystem operations fail on relative paths when cwd()"""
    p = UPath("memory:///foo/bar/baz.txt")
    root = UPath("memory:///foo")
    rel = p.relative_to(root)

    # Memory filesystem doesn't implement cwd(), so these should fail
    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "fsspec paths can not be relative and MemoryPath.cwd() is unsupported"
        ),
    ):
        _ = rel.path

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "fsspec paths can not be relative and MemoryPath.cwd() is unsupported"
        ),
    ):
        rel.exists()


def test_filesystem_operations_work_with_cwd():
    """Test that filesystem operations work on relative paths when cwd()"""
    with tempfile.TemporaryDirectory() as tmpdir:
        # Create test file structure
        test_dir = os.path.join(tmpdir, "testdir")
        os.makedirs(test_dir, exist_ok=True)
        test_file = os.path.join(test_dir, "testfile.txt")
        with open(test_file, "w") as f:
            f.write("test content")

        # Create paths
        abs_path = UPath(test_file)
        abs_dir = UPath(test_dir)
        rel_path = abs_path.relative_to(abs_dir)

        assert not rel_path.is_absolute()
        assert str(rel_path) == "testfile.txt"

        # Change to the test directory and try filesystem operations
        old_cwd = os.getcwd()
        try:
            os.chdir(test_dir)

            # These should work now since we're in the right directory
            full_path = rel_path.path
            assert "testfile.txt" in full_path

            # Test that the file exists
            assert rel_path.exists()

        finally:
            os.chdir(old_cwd)


def test_pickling_relative_paths():
    """Test that relative paths can be pickled and unpickled."""
    p = UPath("memory:///foo/bar/baz.txt")
    root = UPath("memory:///foo")
    rel = p.relative_to(root)

    # Pickle and unpickle
    pickled = pickle.dumps(rel)
    unpickled = pickle.loads(pickled)

    assert str(rel) == str(unpickled)
    assert rel.is_absolute() == unpickled.is_absolute()
    assert rel._relative_base == unpickled._relative_base


def test_with_segments_preserves_relative_state():
    """Test that with_segments preserves the relative state."""
    p = UPath("memory:///foo/bar/baz.txt")
    root = UPath("memory:///foo")
    rel = p.relative_to(root)

    # Create new path with different segments
    new_rel = rel.with_segments("memory:///foo/other/file.txt")

    # Should still be relative with same root
    assert not new_rel.is_absolute()
    assert new_rel._relative_base == rel._relative_base


def test_relative_path_parts():
    """Test that parts work correctly for relative paths."""
    p = UPath("memory:///foo/bar/baz/qux.txt")
    root = UPath("memory:///foo")
    rel = p.relative_to(root)

    assert p.parts == root.parts + rel.parts


def test_absolute_method_behavior():
    """Test that absolute() returns the original absolute path."""
    p = UPath("memory:///foo/bar/baz.txt")
    root = UPath("memory:///foo")
    rel = p.relative_to(root)

    with pytest.raises(
        NotImplementedError,
        match=re.escape("MemoryPath.cwd() is unsupported"),
    ):
        rel.absolute()


def test_is_absolute_method():
    """Test is_absolute() method on relative paths."""
    p = UPath("memory:///foo/bar/baz.txt")
    root = UPath("memory:///foo")
    rel = p.relative_to(root)

    assert not rel.is_absolute()


def test_relative_path_comparison():
    """Test that relative paths can be compared."""
    p1 = UPath("memory:///foo/bar/baz.txt")
    p2 = UPath("memory:///foo/bar/qux.txt")
    root = UPath("memory:///foo")

    rel1 = p1.relative_to(root)
    rel2 = p2.relative_to(root)

    # Compare string representations since .path requires cwd() for memory://
    assert str(rel1) != str(rel2)
    assert rel1 != rel2

    # Same relative path should be equal
    rel1_copy = p1.relative_to(root)
    assert str(rel1) == str(rel1_copy)
    assert rel1 == rel1_copy

    # Same relative path from different base should be equal
    rel3 = UPath("memory:///a/b/c.txt").relative_to(UPath("memory:///a"))
    rel4 = UPath("file:///x/b/c.txt").relative_to(UPath("file:///x"))

    assert str(rel3) == str(rel4)
    assert rel3 == rel4


def test_nonrelative_path_is_absolute():
    """Test that normal (non-relative) paths return True for is_absolute()."""
    p = UPath("memory:///foo/bar/baz.txt")
    assert p.is_absolute()


def test_s3_relative_paths():
    """Test relative paths work with S3 URLs."""
    p = UPath("s3://test_bucket/dir/file.txt")
    root = UPath("s3://test_bucket")
    rel = p.relative_to(root)

    assert not rel.is_absolute()
    assert str(rel) == "dir/file.txt"


@pytest.fixture
def rel_path():
    p = UPath("memory:///foo/bar/baz.txt")
    root = UPath("memory:///foo")
    yield p.relative_to(root)


def test_relative_path_as_uri(rel_path):
    with pytest.raises(
        ValueError,
        match=f"relative path can't be expressed as a {rel_path.protocol} URI",
    ):
        rel_path.as_uri()


@pytest.mark.parametrize(
    "method_args",
    [
        pytest.param(("absolute", ()), id="absolute"),
        pytest.param(("chmod", (0o777,)), id="chmod"),
        pytest.param(("cwd", ()), id="cwd"),
        pytest.param(("exists", ()), id="exists"),
        pytest.param(("glob", ("*.txt",)), id="glob"),
        pytest.param(("group", ()), id="group"),
        pytest.param(("is_dir", ()), id="is_dir"),
        pytest.param(("is_file", ()), id="is_file"),
        pytest.param(("is_symlink", ()), id="is_symlink"),
        pytest.param(("iterdir", ()), id="iterdir"),
        pytest.param(("lchmod", (0o777,)), id="lchmod"),
        pytest.param(("lstat", ()), id="lstat"),
        pytest.param(("mkdir", ()), id="mkdir"),
        pytest.param(("open", ()), id="open"),
        pytest.param(("owner", ()), id="owner"),
        pytest.param(("read_bytes", ()), id="read_bytes"),
        pytest.param(("read_text", ()), id="read_text"),
        pytest.param(("readlink", ()), id="readlink"),
        pytest.param(("rename", ("a/b/c",)), id="rename"),
        pytest.param(("replace", ("...",)), id="replace"),
        pytest.param(("resolve", ()), id="resolve"),
        pytest.param(("rglob", ("*.txt",)), id="rglob"),
        pytest.param(("rmdir", ()), id="rmdir"),
        pytest.param(("samefile", ("...",)), id="samefile"),
        pytest.param(("stat", ()), id="stat"),
        pytest.param(("symlink_to", ("...",)), id="symlink_to"),
        pytest.param(("touch", ()), id="touch"),
        pytest.param(("unlink", ()), id="unlink"),
        pytest.param(("write_bytes", (b"data",)), id="write_bytes"),
        pytest.param(("write_text", ("data",)), id="write_text"),
    ],
)
def test_path_operations_disabled_without_cwd(rel_path, method_args):
    """UPaths without .cwd() implementation should not allow path operations."""
    method, args = method_args

    with (
        pytest.raises(NotImplementedError),
        warnings.catch_warnings(),
    ):
        # warnings are asserted in other tests
        warnings.simplefilter("ignore", UserWarning)
        # next only needs to be called for iterdir and glob/rglob
        # but the other raise already in the getattr call
        next(getattr(rel_path, method)(*args))


@pytest.mark.parametrize(
    "protocol,path,base",
    [
        ("", "/foo/bar/baz.txt", "/foo"),
        ("file", "/foo/bar/baz.txt", "/foo"),
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo"),
        ("ftp", "ftp://user:pass@host/foo/bar/baz.txt", "ftp://user:pass@host/foo"),
        ("http", "http://host/foo/bar/baz.txt", "http://host/foo"),
        ("https", "https://host/foo/bar/baz.txt", "https://host/foo"),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo"),
    ],
)
def test_drive_root_anchor_empty_for_relative_paths(protocol, path, base):
    rel = UPath(path, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    assert (rel.drive, rel.root, rel.anchor) == ("", "", "")


@pytest.mark.parametrize(
    "protocol,path,base,expected_rel",
    [
        ("", "/foo/bar/baz.txt", "/foo", "bar/baz.txt"),
        ("file", "/foo/bar/baz.txt", "/foo", "bar/baz.txt"),
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo", "bar/baz.txt"),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo", "bar/baz.txt"),
        (
            "ftp",
            "ftp://user:pass@host/foo/bar/baz.txt",
            "ftp://user:pass@host/foo",
            "bar/baz.txt",
        ),
        ("http", "http://host/foo/bar/baz.txt", "http://host/foo", "bar/baz.txt"),
        ("https", "https://host/foo/bar/baz.txt", "https://host/foo", "bar/baz.txt"),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo", "bar/baz.txt"),
    ],
)
def test_relative_path_properties(protocol, path, base, expected_rel):
    rel = UPath(path, protocol=protocol).relative_to(UPath(base, protocol=protocol))

    assert not rel.is_absolute()
    assert rel.as_posix() == expected_rel
    assert rel.parts == tuple(expected_rel.split("/"))


@pytest.mark.parametrize(
    "protocol,path,base,expected_parts",
    [
        ("", "/foo/bar/baz.txt", "/foo", ("bar", "baz.txt")),
        ("file", "/foo/bar/baz.txt", "/foo", ("bar", "baz.txt")),
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo", ("bar", "baz.txt")),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo", ("bar", "baz.txt")),
        (
            "ftp",
            "ftp://user:pass@host/foo/bar/baz.txt",
            "ftp://user:pass@host/foo",
            ("bar", "baz.txt"),
        ),
        ("http", "http://host/foo/bar/baz.txt", "http://host/foo", ("bar", "baz.txt")),
        (
            "https",
            "https://host/foo/bar/baz.txt",
            "https://host/foo",
            ("bar", "baz.txt"),
        ),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo", ("bar", "baz.txt")),
    ],
)
def test_relative_path_parts_property(protocol, path, base, expected_parts):
    rel = UPath(path, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    assert rel.parts == expected_parts


def test_relative_path_is_something(rel_path):
    assert rel_path.is_block_device() is False
    assert rel_path.is_char_device() is False
    assert rel_path.is_fifo() is False
    assert rel_path.is_mount() is False
    assert rel_path.is_reserved() is False
    assert rel_path.is_socket() is False


def test_relative_path_hashable():
    x = UPath("memory:///a/b/c.txt")
    y = x.relative_to(UPath("memory:///a"))
    assert hash(y) != hash(x)


def test_relative_path_expanduser_noop(rel_path):
    # this should be revisited if we ever add ~ support to non-file protocols
    assert rel_path == rel_path.expanduser()


def test_relative_path_stem_suffix_name(rel_path):
    assert rel_path.name == "baz.txt"
    assert rel_path.stem == "baz"
    assert rel_path.suffix == ".txt"
    assert rel_path.suffixes == [".txt"]
    assert rel_path.with_name("other.txt").name == "other.txt"
    assert rel_path.with_stem("other").name == "other.txt"
    assert rel_path.with_suffix(".md").name == "baz.md"
    assert rel_path.with_suffix(".tar.gz").suffixes == [".tar", ".gz"]


@pytest.mark.parametrize(
    "protocol,pth,base,expected_parent",
    [
        ("", "/foo/bar/baz.txt", "/foo", "bar"),
        ("", "/foo", "/foo", "."),
        ("file", "/foo/bar/baz.txt", "/foo", "bar"),
        ("file", "/foo/bar", "/foo/bar", "."),
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo", "bar"),
        ("s3", "s3://bucket/foo/bar/", "s3://bucket/foo/bar", "."),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo", "bar"),
        ("gcs", "gcs://bucket/foo/bar/", "gcs://bucket/foo", "."),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo", "bar"),
        ("memory", "memory:///foo/bar", "memory:///foo", "."),
        ("https", "https://host/foo/bar/baz.txt", "https://host/foo", "bar"),
        ("https", "https://host/foo/bar/", "https://host/foo/bar", "."),
    ],
)
def test_relative_path_parent(protocol, pth, base, expected_parent):
    rel = UPath(pth, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    assert str(rel.parent) == expected_parent


@pytest.mark.parametrize(
    "uri,base,expected_parents_parts",
    [
        ("/foo/bar/baz/qux.txt", "/foo", [("bar", "baz"), ("bar",), ()]),
        ("file:///foo/bar/baz/qux.txt", "file:///foo", [("bar", "baz"), ("bar",), ()]),
        ("s3://bucket/foo/bar/baz/", "s3://bucket/", [("foo", "bar"), ("foo",), ()]),
        ("gcs://bucket/foo/bar/baz", "gcs://bucket/", [("foo", "bar"), ("foo",), ()]),
        ("az://bucket/foo/bar/baz", "az://bucket/", [("foo", "bar"), ("foo",), ()]),
        (
            "memory:///foo/bar/baz/qux.txt",
            "memory:///foo",
            [("bar", "baz"), ("bar",), ()],
        ),
        (
            "https://host.com/foo/bar/baz/qux.txt",
            "https://host.com/foo",
            [("bar", "baz"), ("bar",), ()],
        ),
    ],
)
def test_relative_path_parents(uri, base, expected_parents_parts):
    rel = UPath(uri).relative_to(UPath(base))
    parents = list(rel.parents)
    assert [x.parts for x in parents] == expected_parents_parts


@pytest.mark.parametrize(
    "protocol,pth,base",
    [
        ("", "/foo/bar/baz.txt", "/foo"),
        ("file", "/foo/bar/baz.txt", "/foo"),
    ],
)
def test_home_works_for_local_paths(protocol, pth, base):
    rel = UPath(pth, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    prefix = f"{protocol}://" if protocol else ""
    assert rel.home().as_posix() == prefix + UPath.home().as_posix()


@pytest.mark.parametrize(
    "protocol,pth,base",
    [
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo"),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo"),
        ("https", "https://host/foo/bar/baz.txt", "https://host/foo"),
    ],
)
def test_home_raises_for_non_local_paths(protocol, pth, base):
    rel = UPath(pth, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    with pytest.raises(
        NotImplementedError,
        match=re.escape(f"{type(rel).__name__}.home() is unsupported"),
    ):
        rel.home()


@pytest.mark.parametrize(
    "protocol,pth,base",
    [
        ("", "/foo/bar/baz.txt", "/foo"),
        ("file", "/foo/bar/baz.txt", "/foo"),
        ("s3", "s3://bucket/foo/bar/baz.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz.txt", "gcs://bucket/foo"),
        ("memory", "memory:///foo/bar/baz.txt", "memory:///foo"),
        ("https", "https://host/foo/bar/baz.txt", "https://host/foo"),
    ],
)
def test_parser_attribute_available(protocol, pth, base):
    rel_path = UPath(pth, protocol=protocol).relative_to(UPath(base, protocol=protocol))
    assert rel_path.parser is not None


@pytest.mark.parametrize(
    "protocol",
    [
        "",
        "file",
    ],
)
def test_relpath_path_resolve(tmp_path, protocol, monkeypatch):
    """This should work for all path types that support .cwd()"""
    base = UPath(tmp_path, protocol=protocol)
    (base / "a" / "b").mkdir(parents=True)
    (base / "a" / "b" / "file.txt").write_text("data")
    monkeypatch.chdir(base)

    rel = UPath("/xyz/a/b/c/d/../../file.txt", protocol=protocol).relative_to(
        UPath("/xyz", protocol=protocol)
    )

    assert rel.as_posix() == "a/b/c/d/../../file.txt"

    resolved = rel.resolve()
    prefix = f"{protocol}://" if protocol else ""
    assert (
        resolved.as_posix() == prefix + (tmp_path / "a" / "b" / "file.txt").as_posix()
    )
    assert resolved.read_text() == "data"
    assert resolved.is_absolute()
    assert resolved.exists()


@pytest.mark.parametrize(
    "protocol,path,base",
    [
        ("", "/foo/bar/baz/qux.txt", "/foo"),
        ("file", "/foo/bar/baz/qux.txt", "/foo"),
        ("s3", "s3://bucket/foo/bar/baz/qux.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz/qux.txt", "gcs://bucket/foo"),
        ("memory", "memory:///foo/bar/baz/qux.txt", "memory:///foo"),
        ("https", "https://host/foo/bar/baz/qux.txt", "https://host/foo"),
    ],
)
def test_relative_path_match(protocol, path, base):
    """Test that match works correctly for relative paths."""
    rel = UPath(path, protocol=protocol).relative_to(UPath(base, protocol=protocol))

    assert rel.as_posix() == "bar/baz/qux.txt"

    # Should match patterns that match the relative path
    assert rel.match("bar/baz/qux.txt")
    assert rel.match("*/baz/qux.txt")
    assert rel.match("bar/*/qux.txt")
    assert rel.match("*/**/*.txt")  # ** acts like *

    # Should not match patterns that don't match
    assert not rel.match("foo/baz/qux.txt")
    assert not rel.match("*.py")
    assert not rel.match("other.txt")


@pytest.mark.parametrize(
    "protocol,path,base",
    [
        ("", "/foo/bar/baz/qux.txt", "/foo"),
        ("file", "/foo/bar/baz/qux.txt", "/foo"),
        ("s3", "s3://bucket/foo/bar/baz/qux.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz/qux.txt", "gcs://bucket/foo"),
        ("memory", "memory:///foo/bar/baz/qux.txt", "memory:///foo"),
        ("https", "https://host/foo/bar/baz/qux.txt", "https://host/foo"),
    ],
)
def test_relative_path_joinpath(protocol, path, base):
    """Test that joinpath works correctly for relative paths."""
    rel = UPath(path, protocol=protocol).relative_to(UPath(base, protocol=protocol))

    # Test joining with a single segment
    assert rel.as_posix() == "bar/baz/qux.txt"
    joined = rel.joinpath("extra.txt")
    assert joined.as_posix() == "bar/baz/qux.txt/extra.txt"
    assert not joined.is_absolute()

    # Test joining with multiple segments
    joined_multi = rel.joinpath("dir", "file.py")
    assert joined_multi.as_posix() == "bar/baz/qux.txt/dir/file.py"
    assert not joined_multi.is_absolute()

    # Test that the result is still relative with same base
    assert joined.protocol == joined_multi.protocol == protocol
    assert joined.storage_options == joined_multi.storage_options == rel.storage_options


@pytest.mark.parametrize(
    "protocol,path,base",
    [
        ("", "/foo/bar/baz/qux.txt", "/foo"),
        ("file", "/foo/bar/baz/qux.txt", "/foo"),
        ("s3", "s3://bucket/foo/bar/baz/qux.txt", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar/baz/qux.txt", "gcs://bucket/foo"),
        ("memory", "memory:///foo/bar/baz/qux.txt", "memory:///foo"),
        ("https", "https://host/foo/bar/baz/qux.txt", "https://host/foo"),
    ],
)
def test_join_local_absolute_path_to_relative(protocol, path, base, tmp_path):
    """Test that joining an absolute path to a relative path works correctly."""
    rel = UPath(path, protocol=protocol).relative_to(base)

    assert rel.as_posix() == "bar/baz/qux.txt"
    tmp_path.joinpath("bar/baz/qux.txt").parent.mkdir(parents=True, exist_ok=True)
    tmp_path.joinpath("bar/baz/qux.txt").write_text("data")

    assert UPath(tmp_path).joinpath(rel).read_text() == "data"


@pytest.mark.parametrize(
    "protocol,path",
    [
        ("", "/foo/bar"),
        ("file", "/foo/bar"),
        ("s3", "s3://bucket/foo/bar"),
        ("gcs", "gcs://bucket/foo/bar"),
        ("memory", "memory:///foo/bar"),
        ("https", "https://host/foo/bar"),
    ],
)
def test_join_fsspec_absolute_path_to_relative(protocol, path):
    p = UPath(path, protocol=protocol)

    x = p.joinpath(Path("a/b/c").as_posix())
    assert x.path.endswith("foo/bar/a/b/c")


@pytest.mark.parametrize(
    "proto0,path0",
    [
        ("", "/foo/bar"),
        ("file", "/foo/bar"),
        ("s3", "s3://bucket/foo/bar"),
        ("gcs", "gcs://bucket/foo/bar"),
        ("memory", "memory:///foo/bar"),
        ("https", "https://host/foo/bar"),
    ],
)
@pytest.mark.parametrize(
    "proto1,path1,base1",
    [
        ("", "/foo/bar", "/foo"),
        ("file", "/foo/bar", "/foo"),
        ("s3", "s3://bucket/foo/bar", "s3://bucket/foo"),
        ("gcs", "gcs://bucket/foo/bar", "gcs://bucket/foo"),
        ("memory", "memory:///foo/bar", "memory:///foo"),
        ("https", "https://host/foo/bar", "https://host/foo"),
    ],
)
def test_join_fsspec_absolute_path_to_fsspec_relative(
    proto0, path0, proto1, path1, base1
):
    p0 = UPath(path0, protocol=proto0)
    p1 = UPath(path1, protocol=proto1).relative_to(base1)
    assert str(p1) == "bar"

    x = p0.joinpath(p1)
    assert x.path.endswith("foo/bar/bar")



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from upath.core import UPath
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import HDFSStorageOptions

__all__ = ["HDFSPath"]


class HDFSPath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["hdfs"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[HDFSStorageOptions],
        ) -> None: ...

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False
    ) -> None:
        if not exist_ok and self.exists():
            raise FileExistsError(str(self))
        super().mkdir(mode=mode, parents=parents, exist_ok=exist_ok)



"""Pathlib API extended to use fsspec backends."""

from __future__ import annotations

from typing import TYPE_CHECKING

try:
    from upath._version import __version__
except ImportError:
    __version__ = "not-installed"

if TYPE_CHECKING:
    from upath.core import UnsupportedOperation
    from upath.core import UPath

__all__ = ["UPath", "UnsupportedOperation"]


def __getattr__(name):
    if name == "UPath":
        from upath.core import UPath

        globals()["UPath"] = UPath
        return UPath
    elif name == "UnsupportedOperation":
        from upath.core import UnsupportedOperation

        globals()["UnsupportedOperation"] = UnsupportedOperation
        return UnsupportedOperation
    else:
        raise AttributeError(f"module {__name__} has no attribute {name}")



"""pathlib_abc exports for compatibility with pathlib."""

from pathlib_abc import JoinablePath
from pathlib_abc import PathInfo
from pathlib_abc import PathParser
from pathlib_abc import ReadablePath
from pathlib_abc import WritablePath
from pathlib_abc import vfsopen
from pathlib_abc import vfspath

__all__ = [
    "JoinablePath",
    "ReadablePath",
    "WritablePath",
    "PathInfo",
    "PathParser",
    "vfsopen",
    "vfspath",
]



from __future__ import annotations

import os
import warnings
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sequence
from datetime import datetime
from stat import S_IFDIR
from stat import S_IFLNK
from stat import S_IFREG
from typing import Any

__all__ = [
    "UPathStatResult",
]


def _convert_value_to_timestamp(value: Any) -> int | float:
    """Try to convert a datetime-like value to a timestamp."""
    if isinstance(value, (int, float)):
        return value
    elif isinstance(value, str):
        if len(value) == 14:
            return datetime.strptime(value, r"%Y%m%d%H%M%S").timestamp()
        if value.endswith("Z"):
            value = value[:-1] + "+00:00"
        return datetime.fromisoformat(value).timestamp()
    elif isinstance(value, datetime):
        return value.timestamp()
    else:
        warnings.warn(
            f"Cannot convert {value!r} of type {type(value)!r} to a timestamp."
            " Please report this at: https://github.com/fsspec/universal_path/issues",
            RuntimeWarning,
            stacklevel=2,
        )
        raise TypeError(f"Cannot convert {value!r} to a timestamp.")


def _get_stat_result_extra_fields() -> tuple[str, ...]:
    """retrieve the extra fields of the os.stat_result class."""
    # Note:
    #  The lines below let us provide a dictionary with the additional
    #  named fields of the stat_result class as keys and the internal
    #  index of the field as value.
    sr = os.stat_result(range(os.stat_result.n_fields))
    rd = sr.__reduce__()
    assert isinstance(rd, tuple), "unexpected return os.stat_result.__reduce__"
    _, (_, extra) = rd
    extra_fields = sorted(extra, key=extra.__getitem__)
    return tuple(extra_fields)


class UPathStatResult:
    """A stat_result compatible class wrapping fsspec info dicts.

    **Note**: It is unlikely that you will ever have to instantiate
      this class directly. If you want to convert and info dict,
      use: `UPathStatResult.from_info(info)`

    This object may be accessed either as a tuple of
      (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime)
    or via the attributes st_mode, st_ino, st_dev, st_nlink, st_uid, and so on.

    There's an additional method `as_info()` for accessing the info dict.
    This is useful to access additional information provided by the file system
    implementation, that's not covered by the stat_result tuple.

    """

    __slots__ = ("_seq", "_info")
    # Note:
    #   can't derive from os.stat_result at all, and can't derive from
    #   tuple and have slots. So we duck type the os.stat_result class

    # Add the fields and "extra fields" of the os.stat_result class
    _fields = (
        "st_mode",
        "st_ino",
        "st_dev",
        "st_nlink",
        "st_uid",
        "st_gid",
        "st_size",
        "st_atime",
        "st_mtime",
        "st_ctime",
    )
    _fields_extra = _get_stat_result_extra_fields()

    # Provide the n_ attributes of the os.stat_result class for compatibility
    n_sequence_fields = len(_fields)
    n_fields = len(_fields) + len(_fields_extra)
    n_unnamed_fields = len(set(_fields_extra).intersection(_fields))

    if (
        n_fields != os.stat_result.n_fields
        or n_sequence_fields != os.stat_result.n_sequence_fields
        or n_unnamed_fields != os.stat_result.n_unnamed_fields
    ):
        warnings.warn(
            "UPathStatResult: The assumed number of fields in the"
            " stat_result class is not correct. Got: "
            f" {_fields!r}, {_fields_extra!r}, {os.stat_result.n_fields}"
            " This might cause problems? Please report this issue at:"
            " https://github.com/fsspec/universal_path/issues",
            RuntimeWarning,
            stacklevel=2,
        )

    def __init__(
        self,
        stat_result_seq: Sequence[int],
        info_dict: Mapping[str, Any] | None = None,
    ) -> None:
        """init compatible with os.stat_result

        Use `UPathStatResult.from_info(info)` to instantiate from a fsspec info.
        """
        seq = tuple(stat_result_seq)
        if n := len(seq) < self.n_sequence_fields:
            raise TypeError(
                f"{self.__name__} takes at least {self.n_fields}-sequence"
                " ({n}-sequence given)"
            )
        elif n > self.n_fields:
            raise TypeError(
                f"{self.__name__} takes at most {self.n_fields}-sequence"
                " ({n}-sequence given)"
            )
        elif self.n_sequence_fields <= n < self.n_sequence_fields:
            warnings.warn(
                "UPathStatResult: The seq provided more than"
                f" {self.n_sequence_fields} items. Ignoring the extra items...",
                UserWarning,
                stacklevel=2,
            )
        self._seq = seq[: self.n_sequence_fields]
        self._info = info_dict or {}

    def __repr__(self):
        cls_name = type(self).__name__
        seq_attrs = ", ".join(map("{0[0]}={0[1]}".format, zip(self._fields, self)))
        return f"{cls_name}({seq_attrs}, info={self._info!r})"

    def __eq__(self, other):
        if not isinstance(other, UPathStatResult):
            return NotImplemented
        else:
            return self._info == other._info

    # --- access to the fsspec info dict ------------------------------

    @classmethod
    def from_info(cls, info: Mapping[str, Any]) -> UPathStatResult:
        """Create a UPathStatResult from a fsspec info dict."""
        # fill all the fallback default values with 0
        defaults = [0] * cls.n_sequence_fields
        return cls(defaults, info)

    def as_info(self) -> Mapping[str, Any]:
        """Return the fsspec info dict."""
        return self._info

    # --- guaranteed fields -------------------------------------------

    @property
    def st_mode(self) -> int:
        """protection bits"""
        mode = self._info.get("mode")
        if isinstance(mode, int):
            return mode
        elif isinstance(mode, str):
            try:
                return int(mode, 8)
            except ValueError:
                pass

        type_ = self._info.get("type")
        if type_ == "file":
            return S_IFREG  # see: stat.S_ISREG
        elif type_ == "directory":
            return S_IFDIR  # see: stat.S_ISDIR

        if self._info.get("isLink"):
            return S_IFLNK  # see: stat.S_ISLNK

        return self._seq[0]

    @property
    def st_ino(self) -> int:
        """inode"""
        ino = self._info.get("ino")
        if isinstance(ino, int):
            return ino
        return self._seq[1]

    @property
    def st_dev(self) -> int:
        """device"""
        dev = self._info.get("dev")
        if isinstance(dev, int):
            return dev
        return self._seq[2]

    @property
    def st_nlink(self) -> int:
        """number of hard links"""
        nlink = self._info.get("nlink")
        if isinstance(nlink, int):
            return nlink
        return self._seq[3]

    @property
    def st_uid(self) -> int:
        """user ID of owner"""
        for key in ["uid", "owner", "uname", "unix.owner"]:
            try:
                return int(self._info[key])
            except (ValueError, TypeError, KeyError):
                pass
        return self._seq[4]

    @property
    def st_gid(self) -> int:
        """group ID of owner"""
        for key in ["gid", "group", "gname", "unix.group"]:
            try:
                return int(self._info[key])
            except (ValueError, TypeError, KeyError):
                pass
        return self._seq[5]

    @property
    def st_size(self) -> int:
        """total size, in bytes"""
        try:
            return int(self._info["size"])
        except (ValueError, TypeError, KeyError):
            return self._seq[6]

    @property
    def st_atime(self) -> int | float:
        """time of last access"""
        for key in ["atime", "time", "last_accessed", "accessTime"]:
            try:
                raw_value = self._info[key]
            except KeyError:
                continue
            try:
                return _convert_value_to_timestamp(raw_value)
            except (TypeError, ValueError):
                pass
        return self._seq[7]

    @property
    def st_mtime(self) -> int | float:
        """time of last modification"""
        for key in [
            "mtime",
            "LastModified",
            "last_modified",
            "timeModified",
            "modificationTime",
            "modified_at",
            "modify",
        ]:
            try:
                raw_value = self._info[key]
            except KeyError:
                continue
            try:
                return _convert_value_to_timestamp(raw_value)
            except (TypeError, ValueError):
                pass
        return self._seq[8]

    @property
    def st_ctime(self) -> int | float:
        """time of last change"""
        try:
            raw_value = self._info["ctime"]
        except KeyError:
            pass
        else:
            try:
                return _convert_value_to_timestamp(raw_value)
            except (TypeError, ValueError):
                pass
        return self._seq[9]

    @property
    def st_birthtime(self) -> int | float:
        """time of creation"""
        for key in [
            "birthtime",
            "created",
            "creation_time",
            "timeCreated",
            "created_at",
        ]:
            try:
                raw_value = self._info[key]
            except KeyError:
                continue
            try:
                return _convert_value_to_timestamp(raw_value)
            except (TypeError, ValueError):
                pass
        raise AttributeError("birthtime")

    @property
    def st_atime_ns(self) -> int:
        """time of last access in nanoseconds"""
        try:
            return int(self._info["atime_ns"])
        except KeyError:
            pass
        atime = self.st_atime
        if isinstance(atime, float):
            return int(atime * 1e9)
        return atime * 1_000_000_000

    @property
    def st_mtime_ns(self) -> int:
        """time of last modification in nanoseconds"""
        try:
            return int(self._info["mtime_ns"])
        except KeyError:
            pass
        mtime = self.st_mtime
        if isinstance(mtime, float):
            return int(mtime * 1e9)
        return mtime * 1_000_000_000

    @property
    def st_ctime_ns(self) -> int:
        """time of last change in nanoseconds"""
        try:
            return int(self._info["ctime_ns"])
        except KeyError:
            pass
        ctime = self.st_ctime
        if isinstance(ctime, float):
            return int(ctime * 1e9)
        return ctime * 1_000_000_000

    # --- extra fields ------------------------------------------------

    def __getattr__(self, item):
        if item in self._fields_extra:
            return 0  # fallback default value
        raise AttributeError(item)

    # --- os.stat_result tuple interface ------------------------------

    def __len__(self) -> int:
        return len(self._fields)

    def __iter__(self) -> Iterator[int]:
        """the sequence interface iterates over the guaranteed fields.

        All values are integers.
        """
        for field in self._fields:
            yield int(getattr(self, field))

    def index(self, value: int, start: int = 0, stop: int | None = None, /) -> int:
        """the sequence interface index method."""
        if stop is None:
            stop = len(self._seq)
        return self._seq.index(value, start, stop)

    def count(self, value: int) -> int:
        """the sequence interface count method."""
        return self._seq.count(value)

    # --- compatibility with the fsspec info dict interface ------------

    def __getitem__(self, item: int | str) -> Any:
        if isinstance(item, str):
            warnings.warn(
                "Access the fsspec info via `.as_info()[key]`",
                DeprecationWarning,
                stacklevel=2,
            )
            return self._info[item]
        # we need to go via the attributes and cast to int
        attr = self._fields[item]
        return int(getattr(self, attr))

    def keys(self):
        """compatibility with the fsspec info dict interface."""
        warnings.warn(
            "Access the fsspec info via `.as_info().keys()`",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._info.keys()

    def values(self):
        """compatibility with the fsspec info dict interface."""
        warnings.warn(
            "Access the fsspec info via `.as_info().values()`",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._info.values()

    def items(self):
        """compatibility with the fsspec info dict interface."""
        warnings.warn(
            "Access the fsspec info via `.as_info().items()`",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._info.items()

    def get(self, key, default=None):
        """compatibility with the fsspec info dict interface."""
        warnings.warn(
            "Access the fsspec info via `.as_info().get(key, default)`",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._info.get(key, default)

    def copy(self):
        """compatibility with the fsspec info dict interface."""
        warnings.warn(
            "Access the fsspec info via `.as_info().copy()`",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._info.copy()
